/** 
 * XMLSec library
 *
 * Keys
 *
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>
 
#include <libxml/tree.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysInternal.h>
#include <xmlsec/transforms.h>
#include <xmlsec/transformsInternal.h>
#include <xmlsec/x509.h>

#include <xmlsec/xmlenc.h>


static xmlSecKeyId xmlSecAllKeyIds[100];

static xmlChar* 	xmlSecKeyNameRead		(const xmlNodePtr keyNameNode);
static int 		xmlSecKeyNameWrite		(xmlNodePtr keyNameNode,
							 xmlSecKeyPtr key);
static xmlSecKeyPtr	xmlSecKeyValueRead		(const xmlNodePtr keyValueNode, 
							 xmlSecTransformId id, 
							 xmlSecKeyType type);
static int 		xmlSecKeyValueWrite		(xmlNodePtr keyValueNode,
							 xmlSecKeyPtr key,
							 xmlSecKeyType type);
static xmlSecKeyPtr	xmlSecRetrievalMethodRead	(const xmlNodePtr retrievalMethodNode,
							 xmlSecTransformId id,
							 xmlSecKeyType type,
							 int level,
							 const xmlSecKeysReadContext *ctx);
static xmlSecKeyPtr	xmlSecKeyNodesListRead		(xmlNodePtr cur, 
							 xmlSecTransformId id, 
							 xmlSecKeyType type, 
							 int level, 
							 const xmlSecKeysReadContext *ctx);

static const xmlChar xmlSecRawX509Cert[] = "http://www.w3.org/2000/09/xmldsig#rawX509Certificate";


/**
 * xmlKeyInfoAddKeyName:
 * @keyInfoNode: the key info node
 *
 *
 *
 */
xmlNodePtr	
xmlKeyInfoAddKeyName(xmlNodePtr keyInfoNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlKeyInfoAddKeyName";
    xmlNodePtr cur;
    
    if((keyInfoNode == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfoNode is null\n",
	    func);	
#endif
	return(NULL);	
    }
        
    cur = xmlSecFindChild(keyInfoNode, BAD_CAST "KeyName", xmlSecDSigNs);
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyName node is already present\n",
	    func);	
#endif
	return(NULL);	
    }
    
    cur = xmlSecAddChild(keyInfoNode, BAD_CAST "KeyName", xmlSecDSigNs); 
    if(cur == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create KeyName node\n",
	    func);	
#endif
	return(NULL);	
    }
    
    return(cur);
}

/**
 * xmlKeyInfoAddKeyValue
 * @keyInfoNode: the key info node
 *
 *
 *
 */
xmlNodePtr
xmlKeyInfoAddKeyValue(xmlNodePtr keyInfoNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlKeyInfoAddKeyValue";
    xmlNodePtr cur;
    
    if((keyInfoNode == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfoNode is null\n",
	    func);	
#endif
	return(NULL);	
    }
        
    cur = xmlSecFindChild(keyInfoNode, BAD_CAST "KeyValue", xmlSecDSigNs);
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyValue node is already present\n",
	    func);	
#endif
	return(NULL);	
    }
    
    cur = xmlSecAddChild(keyInfoNode, BAD_CAST "KeyValue", xmlSecDSigNs); 
    if(cur == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create KeyValue node\n",
	    func);	
#endif
	return(NULL);	
    }
    
    return(cur);
}

/**
 * xmlKeyInfoAddX509Data:
 * @keyInfoNode: the key info node
 *
 *
 *
 */
xmlNodePtr
xmlKeyInfoAddX509Data(xmlNodePtr keyInfoNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlKeyInfoAddX509Data";
    xmlNodePtr cur;
    
    if((keyInfoNode == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfoNode is null\n",
	    func);	
#endif
	return(NULL);	
    }
        
    cur = xmlSecFindChild(keyInfoNode, BAD_CAST "X509Data", xmlSecDSigNs);
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509Data node is already present\n",
	    func);	
#endif
	return(NULL);	
    }
    
    cur = xmlSecAddChild(keyInfoNode, BAD_CAST "X509Data", xmlSecDSigNs); 
    if(cur == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create X509Data node\n",
	    func);	
#endif
	return(NULL);	
    }
    
    return(cur);
}

/**
 * xmlKeyInfoAddRetrievalMethod
 * @keyInfoNode: the key info node
 * @uri: the reference uri
 * @type:
 *
 */
xmlNodePtr
xmlKeyInfoAddRetrievalMethod(xmlNodePtr keyInfoNode, const xmlChar *uri,
			     const xmlChar *type) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlKeyInfoAddRetrievalMethod";
    xmlNodePtr cur;
    
    if((keyInfoNode == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfoNode is null\n",
	    func);	
#endif
	return(NULL);	
    }
        
    cur = xmlSecFindChild(keyInfoNode, BAD_CAST "RetrievalMethod", xmlSecDSigNs);
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: RetrievalMethod node is already present\n",
	    func);	
#endif
	return(NULL);	
    }
    
    cur = xmlSecAddChild(keyInfoNode, BAD_CAST "RetrievalMethod", xmlSecDSigNs); 
    if(cur == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create RetrievalMethod node\n",
	    func);	
#endif
	return(NULL);	
    }
    
    if(uri != NULL) {
	xmlSetProp(cur, BAD_CAST "URI", uri);
    }

    if(type != NULL) {
	xmlSetProp(cur, BAD_CAST "Type", type);
    }
    return(cur);
}

/**
 * 
 * @keyInfoNode: the key info node
 *
 *
 *
 */
xmlNodePtr
xmlKeyInfoAddRetrievalMethodTransform(xmlNodePtr keyInfoNode,
			     xmlSecTransformId transform) {
    static const char func[] _UNUSED_VARIABLE_ = "";
    xmlNodePtr retrMethod;
    xmlNodePtr transforms;
    xmlNodePtr cur;
    int ret;
    
    if((keyInfoNode == NULL) || (transform == xmlSecTransformUnknown)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfoNode is null or transform is unknown\n",
	    func);	
#endif
	return(NULL);	
    }
        
    retrMethod = xmlSecFindChild(keyInfoNode, BAD_CAST "RetrievalMethod", xmlSecDSigNs);
    if(retrMethod == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: RetrievalMethod node is not found\n",
	    func);	
#endif
	return(NULL);	
    }
    
    transforms = xmlSecFindChild(retrMethod, BAD_CAST "Transforms", xmlSecDSigNs);
    if(transforms == NULL) {
	transforms = xmlSecAddChild(retrMethod, BAD_CAST "Transforms", xmlSecDSigNs);
	if(transforms == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create Transforms node\n", 
		func);	
#endif
	    return(NULL);	
	}
    }
    
    
    cur = xmlSecAddChild(transforms, BAD_CAST "Transform", xmlSecDSigNs); 
    if(cur == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create Transform node\n",
	    func);	
#endif
	return(NULL);	
    }
    
    ret = xmlSecTransformNodeWrite(cur, transform);
    if(ret < 0){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: transform write failed\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    return(cur);
}


/**
 * xmlSecKeyInit
 * 
 * Initializes the keys list
 */
void		
xmlSecKeysInit(void) {
    int i = 0;

#ifndef XMLSEC_NO_HMAC    
    xmlSecAllKeyIds[i++] = xmlSecHmacKey;
#endif /* XMLSEC_NO_HMAC */    

#ifndef XMLSEC_NO_DSA
    xmlSecAllKeyIds[i++] = xmlSecDsaKey;
#endif /* XMLSEC_NO_DSA */    

#ifndef XMLSEC_NO_RSA
    xmlSecAllKeyIds[i++] = xmlSecRsaKey;
#endif /* XMLSEC_NO_RSA */

#ifndef XMLSEC_NO_DES    
    xmlSecAllKeyIds[i++] = xmlSecDesKey;
#endif /* XMLSEC_NO_DES */

#ifndef XMLSEC_NO_AES    
    xmlSecAllKeyIds[i++] = xmlSecAesKey;
#endif /* XMLSEC_NO_AES */

    /* MUST be the last in the list */  
    xmlSecAllKeyIds[i++] = xmlSecKeyIdUnknown;
}

/**
 * xmlSecKeySearcCtxFindKey:
 * @keysContext:
 * @id:
 * @type:
 * @name:
 *
 */
xmlSecKeyPtr 	
xmlSecKeySearcCtxFindKey(const xmlSecKeysReadContext *keysContext,
			 xmlSecTransformId id, xmlSecKeyType type,
			 const xmlChar *name) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeySearcCtxFindKey";
    xmlSecKeyPtr key = NULL;
    xmlSecKeyId keyId;
    
    if(keysContext == NULL){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key search context is null\n",
	    func);	
#endif
	return(NULL);	
    }
    
    if(id == xmlSecTransformUnknown) {
	keyId = xmlSecKeyIdUnknown;
    } else if(id->type == xmlSecTransformTypeBinary) {
	keyId = ((xmlSecBinTransformId)id)->keyId;
    } else {	
	return(NULL);
    }
				    
    if(keysContext->findKeyCallback != NULL) {
	key = (keysContext->findKeyCallback)(keysContext->findKeyContext,
					     keyId, type, name);
    }
    return(key);
}

/**
 * xmlSecKeyCreate
 * @id:
 * @origin:
 *
 * Creates new key (wrapper for xmlSecKey::create method)
 *
 */
xmlSecKeyPtr	
xmlSecKeyCreate(xmlSecKeyId id, xmlSecKeyOrigin origin)  {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyCreate";
    xmlSecKeyPtr key;
    
    if((id == xmlSecKeyIdUnknown) || (id->create == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: id or create method is not defined\n",
	    func);	
#endif
	return(NULL);	
    }
    
    key = id->create(id);
    if(key == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key creation failed\n");	
#endif
	return(NULL);	
    }
    key->origin = origin;
    return(key);
}

/**
 * xmlSecKeyDestroy
 * @key:
 * @forceDestroy:
 *
 * Destroys the key. If the key origin has a KeyManager marker 
 * the key is *not* destroyed unless @forceDestroy is not zero
 */
void
xmlSecKeyDestroy(xmlSecKeyPtr key,  int forceDestroy) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyDestroy";
    
    if((!xmlSecKeyIsValid(key)) || (key->id->destroy == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or destroy method is not defined\n",
	    func);	
#endif
	return;
    }
    
    if((key->origin & xmlSecKeyOriginKeyManager) && (!forceDestroy)) {
	/* 
	 * we do not destroy keys from key manager unless it is
	 * specially requested
	 */
	return;
    }
    if(key->name != NULL) {
	xmlFree(key->name); 
	key->name = NULL;
    }
#ifndef XMLSEC_NO_X509
    if(key->x509Chain != NULL) {	
	sk_X509_pop_free(key->x509Chain, X509_free); 
	key->x509Chain = NULL;
    } else if(key->x509Cert != NULL) {	
	X509_free(key->x509Cert); 
	key->x509Cert = NULL;
    }
#endif /* XMLSEC_NO_X509 */    
    key->id->destroy(key);
}

/**
 * xmlSecKeyReadXml
 * @node:
 * 
 * Reads the key form XML doc
 */
xmlSecKeyPtr	
xmlSecKeyReadXml(xmlSecKeyId id, xmlNodePtr node) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyReadXml";
    xmlSecKeyPtr key;
    int ret;
    
    if(node == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: node is null\n",
	    func);	
#endif
	return(NULL);
    }

    if(id->read == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: XML read not implemented for key \"%s\"\n",
	    func, id->keyValueNodeName);	
#endif
	return(NULL);	    	
    }

    key = xmlSecKeyCreate(id, xmlSecKeyOriginDefault);
    if(key == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key\n",
	    func);	
#endif
	return(NULL);    
    }

    
    
    ret = (id->read)(key, node);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read key\n",
	    func);	
#endif
	xmlSecKeyDestroy(key, 1);
	return(NULL);    
    }
    
    return(key);
}

/**
 * xmlSecKeyWriteXml
 * @key:
 * @type:
 * @parent:
 * 
 * Writes the key into XML node and adds it to the list of @parent children
 * (if parent is not null).
 */
int
xmlSecKeyWriteXml(xmlSecKeyPtr key, xmlSecKeyType type, xmlNodePtr node) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyWriteXml";
    int ret;
    
    if((!xmlSecKeyIsValid(key)) || (node == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or parent is null\n",
	    func);	
#endif
	return(-1);
    }
    
    if(key->id->write == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: XML write not implemented for key \"%s\"\n",
	    func, key->id->keyValueNodeName);	
#endif
	return(-1);	    	
    }
    
    /* write key */
    ret = key->id->write(key, type, node);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write key\n",
	    func);	
#endif
	return(-1);	    
    }
    return(0);    
}

xmlSecKeyPtr	
xmlSecKeyReadBin(xmlSecKeyId id, const unsigned char *buf, size_t size) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyReadBin";
    xmlSecKeyPtr key;
    int ret;
    
    if((buf == NULL) || (size == 0)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: buffer or size is null\n",
	    func);	
#endif
	return(NULL);
    }

    if(id->readBin == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: bin read not implemented for key \"%s\"\n",
	    func, id->keyValueNodeName);	
#endif
	return(NULL);	    	
    }

    key = xmlSecKeyCreate(id, xmlSecKeyOriginDefault);
    if(key == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key\n",
	    func);	
#endif
	return(NULL);    
    }

    
    
    ret = (id->readBin)(key, buf, size);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read key\n",
	    func);	
#endif
	xmlSecKeyDestroy(key, 1);
	return(NULL);    
    }
    
    return(key);
}

int
xmlSecKeyWriteBin(xmlSecKeyPtr key, xmlSecKeyType type,
		 unsigned char **buf, size_t *size) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyWriteBin";
    int ret;
    
    if((!xmlSecKeyIsValid(key)) || (buf == NULL) || (size == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key is invalid or buffer or size is null\n",
	    func);	
#endif
	return(-1);
    }

    if(key->id->writeBin == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: bin write not implemented for key \"%s\"\n",
	    func, key->id->keyValueNodeName);	
#endif
	return(-1);	    	
    }
    
    
    /* write key */
    ret = key->id->writeBin(key, type, buf, size);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write key\n",
	    func);	
#endif
	return(-1);	    
    }
    return(0);    
}

/** 
 * xmlSecDebugKeyDump
 *
 *
 *
 */
void
xmlSecDebugKeyDump(FILE *output, xmlSecKeyPtr key) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecDebugKeyDump";
    
    if((output == NULL) || !xmlSecKeyIsValid(key)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key or output file is null\n", 
	    func);	
#endif
	return;
    }
    fprintf(output, "== KEY\n");
    fprintf(output, "=== method: %s\n", 
	    (key->id->keyValueNodeName != NULL) ? 
	    (char*)(key->id->keyValueNodeName) : "NULL"); 
    fprintf(output, "=== key name: %s\n", 
	    (key->name != NULL) ? 
	    (char*)(key->name) : "NULL"); 
    fprintf(output, "=== key type: %s\n", 
	    (key->type == xmlSecKeyTypePrivate) ? 
	    "Private" : "Public"); 
    fprintf(output, "=== key origin:");
    if(key->origin & xmlSecKeyOriginKeyManager) {
	fprintf(output, " KeyManager");
    }
    if(key->origin & xmlSecKeyOriginKeyName) {
	fprintf(output, " KeyName");
    }
    if(key->origin & xmlSecKeyOriginKeyValue) {
	fprintf(output, " KeyValue");
    }
    if(key->origin & xmlSecKeyOriginRetrievalDocument) {
	fprintf(output, " RetrievalDocument");
    }
    if(key->origin & xmlSecKeyOriginRetrievalRemote) {
	fprintf(output, " RetrievalRemote");
    }
    if(key->origin & xmlSecKeyOriginX509) {
	fprintf(output, " x509");
    }
    if(key->origin & xmlSecKeyOriginEncryptedKey) {
	fprintf(output, " EncKey");
    }
    if(key->origin & xmlSecKeyOriginPGP) {
	fprintf(output, " PGP");
    }
    fprintf(output, "\n");
#ifndef XMLSEC_NO_X509
    if(key->x509Cert != NULL) {
	xmlSecDebugx509Dump(output, key->x509Cert, key->x509Chain);
    }
#endif /* XMLSEC_NO_X509 */    
}

/** 
 * XMLDsig:KeyInfo methods
 * http://www.w3.org/TR/xmlSec-core/#sec-KeyInfo
 *
 * The KeyInfo Element
 *
 * KeyInfo is an optional element that enables the recipient(s) to obtain 
 * the key needed to validate the signature.  KeyInfo may contain keys, 
 * names, certificates and other public key management information, such as 
 * in-band key distribution or key agreement data. 
 * 
 *  Schema Definition:
 *
 *  <element name="KeyInfo" type="ds:KeyInfoType"/> 
 *  <complexType name="KeyInfoType" mixed="true">
 *    <choice maxOccurs="unbounded"> 
 *       <element ref="ds:KeyName"/> 
 *       <element ref="ds:KeyValue"/> 
 *       <element ref="ds:RetrievalMethod"/> 
 *       <element ref="ds:X509Data"/> 
 *       <element ref="ds:PGPData"/> 
 *       <element ref="ds:SPKIData"/>
 *       <element ref="ds:MgmtData"/>
 *       <any processContents="lax" namespace="##other"/>
 *       <!-- (1,1) elements from (0,unbounded) namespaces -->
 *    </choice>
 *    <attribute name="Id" type="ID" use="optional"/>
 *  </complexType>
 *    
 * DTD:
 *    
 * <!ELEMENT KeyInfo (#PCDATA|KeyName|KeyValue|RetrievalMethod|
 *                    X509Data|PGPData|SPKIData|MgmtData %KeyInfo.ANY;)* >      
 * <!ATTLIST KeyInfo  Id  ID   #IMPLIED >
 *  
 */ 
/**
 * xmlSecKeyInfoRead:
 * @keyInfoNode:
 * @id:			the trasnform we need key for
 * @type:		the type of the key we are looking for
 * @ctx:
 *
 * Reads the key from key info node. We ignore the unknown nodes and hope 
 * for the best
 */
xmlSecKeyPtr	
xmlSecKeyInfoRead(const xmlNodePtr keyInfoNode, xmlSecTransformId id,
	    	xmlSecKeyType type, const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyInfoRead";
    xmlSecKeyPtr key = NULL;
    xmlNodePtr cur;
    
    if((keyInfoNode == NULL) || (ctx == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyInfo node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }
    cur = xmlSecGetNextElementNode(keyInfoNode->children); 
    if(keyInfoNode->children != NULL) {
	key = xmlSecKeyNodesListRead(cur, id, type, 0, ctx);
    }
    return(key);
}

/**
 * xmlSecKeyNodesListRead:
 * @cur:
 * @id:			the trasnform we need key for
 * @type:		the type of the key we are looking for
 * @level:
 * @ctx:
 *
 */
static xmlSecKeyPtr
xmlSecKeyNodesListRead(xmlNodePtr cur, xmlSecTransformId id, xmlSecKeyType type, 
		int level, const xmlSecKeysReadContext *ctx) { 
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyNodesListRead";
    xmlChar *keyName;
    xmlSecKeyPtr key;
    
    if(ctx == NULL){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyInfo node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }
    
    key = NULL;
    keyName = NULL;
    while ((key == NULL) && (cur != NULL)) {
	if(xmlSecCheckNodeName(cur, BAD_CAST "KeyName", xmlSecDSigNs)) {
	    if(keyName != NULL) {
	        xmlFree(keyName);
	    }		
	    keyName = xmlSecKeyNameRead(cur);
	    if((keyName != NULL) && (ctx->allowedOrigins & xmlSecKeyOriginKeyManager) && 
			(ctx->allowedOrigins & xmlSecKeyOriginKeyName)) {
		key = xmlSecKeySearcCtxFindKey(ctx, id, type, keyName);
		if(key != NULL) {
		    xmlFree(keyName);
		    return(key);
		}
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "KeyValue", xmlSecDSigNs) &&
			(ctx->allowedOrigins & xmlSecKeyOriginKeyValue)) {
	    key = xmlSecKeyValueRead(cur, id, type);
	    if((key != NULL) && (keyName != NULL)){
	        key->name = keyName;
	        keyName = NULL;
	        return(key);
	    }
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "RetrievalMethod", xmlSecDSigNs) &&
			(ctx->allowedOrigins & (xmlSecKeyOriginRetrievalDocument |
						xmlSecKeyOriginRetrievalRemote)) &&
			(level < ctx->maxRetrievals)) {
	    key = xmlSecRetrievalMethodRead(cur, id, type, level, ctx);
#ifndef XMLSEC_NO_X509
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509Data", xmlSecDSigNs) && 
			(ctx->allowedOrigins & xmlSecKeyOriginX509)) {
	    key = xmlSecX509DataNodeRead(cur, id, type, &(ctx->x509));
#endif /* XMLSEC_NO_X509 */ 	
#ifndef XMLSEC_NO_XMLENC
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "EncryptedKey", xmlSecEncNs) && 
			(ctx->allowedOrigins & xmlSecKeyOriginEncryptedKey) &&
			(ctx->maxEncKeyLevels > 0)) {
	    key = xmlEncryptedKeyNodeRead(cur, id, type, ctx);
#endif /* XMLSEC_NO_XMLENC */
	}
	/* TODO: add more nodes (pgp, spki, etc) */
	cur = xmlSecGetNextElementNode(cur->next);
    }    
    if(keyName != NULL) {
	xmlFree(keyName);
    }
    return(key);
}

/**
 * xmlSecKeyInfoWrite
 * @keyInfoNode:
 * @key:
 * @type:
 *
 * Writes the key in the key info node according to the KeyInfo node
 * content
 */
int
xmlSecKeyInfoWrite(xmlNodePtr keyInfoNode, const xmlSecKeyPtr key,
		    xmlSecKeyType type, const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyInfoWrite";
    xmlNodePtr cur;
    int ret;
        
    if(keyInfoNode == NULL){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfo node  is null\n",
	    func);	
#endif
	return(-1);
    }

    ret = 0;
    cur = xmlSecGetNextElementNode(keyInfoNode->children);
    while(cur != NULL) {
	if(xmlSecCheckNodeName(cur, BAD_CAST "KeyName", xmlSecDSigNs)) {
	    ret = xmlSecKeyNameWrite(cur, key);
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "KeyValue", xmlSecDSigNs)) {
	    ret = xmlSecKeyValueWrite(cur, key, type);
#ifndef XMLSEC_NO_X509
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "X509Data", xmlSecDSigNs)) {
	    ret = xmlSecX509DataNodeWrite(cur, key); 
#endif /* XMLSEC_NO_X509 */	
#ifndef XMLSEC_NO_XMLENC
	} else if(xmlSecCheckNodeName(cur, BAD_CAST "EncryptedKey", xmlSecEncNs)) {
	    ret = xmlEncryptedKeyNodeWrite(cur, key, type, ctx);
#endif /* XMLSEC_NO_XMLENC */
	}
	/* TODO: add retrieval method, etc. */
	if(ret < 0) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: keyInfo failed for \"%s\"\n",
		func, cur->name);	
#endif
	    return(-1);	    
	}
	cur = xmlSecGetNextElementNode(cur->next);
    }    
    return(0);
}		    


/**
 * xmlSecKeyNameRead:
 * @keyNameNode:
 *
 *
 */
static xmlChar*
xmlSecKeyNameRead(const xmlNodePtr keyNameNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyNameRead";    
    xmlChar *content;
    
    if(keyNameNode == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyName node is null\n",
	    func);	
#endif
	return(NULL);
    }
    
    content = xmlNodeGetContent(keyNameNode);
    /* TODO: decode key name */    
    return(content);
}

/** 
 * xmlSecKeyNameWrite:
 * @keyNameNode:
 * @key:
 *
 *
 */
static int 
xmlSecKeyNameWrite(xmlNodePtr keyNameNode, xmlSecKeyPtr key) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyNameWrite";
    
    if((keyNameNode == NULL) || (!xmlSecKeyIsValid(key))) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyName node or key is not valid\n",
	    func);	
#endif
	return(-1);    
    }
    
    if(key->name != NULL) {
	/* TODO: encode the key name! */
	xmlNodeSetContent(keyNameNode, key->name);
    }
    return(0);
}


/**
 * xmlSecKeyValueRead
 * @keyValueNode:
 * @id:
 * @type:
 *
 * http://www.w3.org/TR/xmldsig-core/#sec-KeyValue
 *
 * The KeyValue element contains a single public key that may be useful in 
 * validating the signature. Structured formats for defining DSA (REQUIRED) 
 * and RSA (RECOMMENDED) public keys are defined in Signature Algorithms 
 * (section 6.4). The KeyValue element may include externally defined 
 * public keys values represented as PCDATA or element types from an external 
 * namespace.
 *
 * Schema Definition:
 *
 * <element name="KeyValue" type="ds:KeyValueType"/> 
 *   <complexType name="KeyValueType" mixed="true">
 *     <choice>
 *       <element ref="ds:DSAKeyValue"/>
 *       <element ref="ds:RSAKeyValue"/>
 *       <any namespace="##other" processContents="lax"/>
 *     </choice>
 *   </complexType>
 *    
 * DTD:
 *  
 * <!ELEMENT KeyValue (#PCDATA|DSAKeyValue|RSAKeyValue %KeyValue.ANY;)* >
 *       
 * =========================================================================
 * Support for private keys is added (@type parameter)
 */
static xmlSecKeyPtr
xmlSecKeyValueRead(const xmlNodePtr keyValueNode, xmlSecTransformId id, 
		xmlSecKeyType type) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyValueRead";   
    xmlNodePtr cur; 
    xmlSecKeyId keyId;
    xmlSecKeyPtr key;
    size_t i;
    
    if(keyValueNode == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyValue node is null\n",
	    func);	
#endif
	return(NULL);
    }
    
    key = NULL;
    cur = xmlSecGetNextElementNode(keyValueNode->children);    
    while(cur != NULL) {
        for(i = 0; xmlSecAllKeyIds[i] != xmlSecKeyIdUnknown; ++i) {
	    keyId = xmlSecAllKeyIds[i];
	    if((id != xmlSecTransformUnknown) && 
	        !xmlSecBinTransformIdCheckKeyId(id, keyId)) continue;
	    if(xmlSecCheckNodeName(cur, keyId->keyValueNodeName, 
					keyId->keyValueNodeNs)) {
		key = xmlSecKeyReadXml(keyId, cur);
		break;
	    }
	}
	if(key != NULL) {
	    if((key->type == type) || (type == xmlSecKeyTypeAny)) {
		return(key);
	    } else {
		xmlSecKeyDestroy(key, 0);
		key = NULL;
	    }
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }
    return(key);
}

/**
 * xmlSecKeyValueWrite:
 * @keyValueNode:
 * @key:
 * @type:
 *
 *
 */
static int 
xmlSecKeyValueWrite(xmlNodePtr keyValueNode, xmlSecKeyPtr key,  xmlSecKeyType type) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyValueWrite";
    xmlNodePtr cur;
    int ret;
    
    if((keyValueNode == NULL) || (!xmlSecKeyIsValid(key))) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: KeyValue node or key is null\n",
	    func);	
#endif
	return(-1);
    }
    
    /* remove all existing key value */
    xmlNodeSetContent(keyValueNode, NULL);
    
    /* create key node */
    cur = xmlSecAddChild(keyValueNode, key->id->keyValueNodeName, key->id->keyValueNodeNs);
    if(cur == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create key node\n",
	    func);	
#endif
	return(-1);	
    }
    
    ret = xmlSecKeyWriteXml(key, type, cur);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write key node\n",
	    func);	
#endif
	xmlUnlinkNode(cur);
	xmlFreeNode(cur);
    }
    return(0);    
}

/**
 *
 *
 *
 *
 *
 */
static xmlSecKeyPtr 		
xmlSecRetrievalMethodRead(const xmlNodePtr retrievalMethodNode,
			 xmlSecTransformId id, xmlSecKeyType type,
			 int level, const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecRetrievalMethodRead";			 
    xmlSecKeyPtr res = NULL;
    xmlNodePtr cur;
    xmlSecTransformStatePtr state = NULL;
    xmlChar *uri = NULL;
    xmlChar *retrType = NULL;
    int ret;
 
    if((retrievalMethodNode == NULL) || (ctx == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: RetrievalMethod node or key search context is null\n",
	    func);	
#endif
	return(NULL);
    }    
    cur = xmlSecGetNextElementNode(retrievalMethodNode->children);
    
    /* read attributes first */
    uri = xmlGetProp(retrievalMethodNode, BAD_CAST "URI");
    if((uri == NULL) || (xmlStrlen(uri) == 0) || (uri[0] == '#')) {
	/* same document uri */
	if((ctx->allowedOrigins & xmlSecKeyOriginRetrievalDocument) == 0) {
	    /* not allowed */
	    xmlFree(uri);
	    return(NULL);
	}
    } else {
	/* remote document */
	if ((ctx->allowedOrigins & xmlSecKeyOriginRetrievalRemote) == 0) {
	    xmlFree(uri);
	    return(NULL);
	}
    }

    state = xmlSecTransformStateCreate(retrievalMethodNode->doc, NULL, (char*)uri);
    if(state == NULL){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create transforms state\n",
	    func);
#endif	    
	goto done;
    }	

    /* first is optional Transforms node */
    if((cur != NULL) && xmlSecCheckNodeName(cur, BAD_CAST "Transforms", xmlSecDSigNs)) {
	ret = xmlSecTransformsNodeRead(state, cur);
	if(ret < 0){
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read \"Transforms\"\n",
		func);
#endif	    
	    goto done;
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }

    ret = xmlSecTransformStateFinal(state, xmlSecTransformResultBinary);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize transforms\n",
	    func);
#endif
	goto done;
    }
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: unexpected node \"%s\" found\n",
	    func, cur->name);
#endif		
	goto done;
    }

    retrType = xmlGetProp(retrievalMethodNode, BAD_CAST "Type");
    if(xmlStrEqual(retrType, xmlSecRawX509Cert)) {
	/* special case: raw DER x509 certificate */
#ifndef XMLSEC_NO_X509
	res = xmlSecDerX509Read((unsigned char*)xmlBufferContent(state->curBuf),
				xmlBufferLength(state->curBuf),
				id, type, &(ctx->x509));
#endif /* XMLSEC_NO_X509 */				
    } else {
	xmlDocPtr keyDoc;
	
        keyDoc = xmlRecoverMemory((char*)xmlBufferContent(state->curBuf),
                              xmlBufferLength(state->curBuf));
	if(keyDoc == NULL) {
#ifdef XMLSEC_DEBUG
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to parse keys doc\n");
#endif
	    xmlFreeDoc(keyDoc);
	    goto done;
	}
        res = xmlSecKeyNodesListRead(xmlDocGetRootElement(keyDoc), 
				id, type, level + 1, ctx);        
	xmlFreeDoc(keyDoc);
    }
    
done:
    if(state != NULL) {
	xmlSecTransformStateDestroy(state);
    }
    if(uri != NULL) {
	xmlFree(uri);
    }
    if(retrType != NULL) {
	xmlFree(retrType);
    }
    return(res);
}


/**
 * xmlSecKeyReadEvpKey:
 * @pKey:
 * @id:
 * @origin:
 *
 * Converts pKey structure to xmlSecKey
 *
 */
xmlSecKeyPtr
xmlSecKeyReadEvpKey(EVP_PKEY *pKey, xmlSecTransformId id, xmlSecKeyOrigin origin) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecKeyReadEvpKey";
    xmlSecKeyPtr key = NULL;
    
    if(pKey == NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: pKey is null\n",
	    func);
#endif	    
	return(NULL);
    }

    switch(pKey->type) {	
#ifndef XMLSEC_NO_RSA    
    case EVP_PKEY_RSA:
	if((id == xmlSecSignRsaSha1) || (id == xmlSecTransformUnknown)) {
	    key = xmlSecKeyCreate(xmlSecRsaKey, origin);
	    if(key != NULL) {
		xmlSecRsaKeyGenerate(key, pKey->pkey.rsa);
		/* hack */
		pKey->pkey.rsa = NULL;
	    }
	}
	break;
#endif /* XMLSEC_NO_RSA */	
#ifndef XMLSEC_NO_DSA	
    case EVP_PKEY_DSA:
	if((id == xmlSecSignDsaSha1) || (id == xmlSecTransformUnknown)) {
	    key = xmlSecKeyCreate(xmlSecDsaKey, origin);
	    if(key != NULL) {
		xmlSecDsaKeyGenerate(key, pKey->pkey.dsa);
		/* hack */
		pKey->pkey.dsa = NULL;
	    }
	}
	break;
#endif /* XMLSEC_NO_DSA */	
    default:	
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the key type %d is not supported\n",
	    func, pKey->type);	
#endif
	return(NULL);
    }
    if(key == NULL) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create xmlSecKey or key not found\n",
	    func);	
#endif
	return(NULL);
    }
    return(key);      
}



