/** 
 * XMLSec library
 *
 * Simple X509 Context Implementation
 *
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#ifndef XMLSEC_NO_X509

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include <openssl/asn1.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/keys.h>
#include <xmlsec/base64.h>
#include <xmlsec/x509.h>

#include <xmlsec/x509mngr.h>

static X509_NAME *	xmlSecX509NameRead		(unsigned char *str, 
							 int len);
static int 		xmlSecX509NameStringRead	(unsigned char **str, 
							 int *strLen, 
							 unsigned char *res, 
							 int resLen, 
							 unsigned char delim, 
							 int ingoreTrailingSpaces);
static int		xmlSecX509NamesCompare		(X509_NAME *a,
							 X509_NAME *b);
static int 		xmlSecX509_NAME_cmp		(const X509_NAME *a, 
							 const X509_NAME *b);
static int 		xmlSecX509_NAME_ENTRY_cmp	(const X509_NAME_ENTRY **a, 
							 const X509_NAME_ENTRY **b);

int			xmlSecSimpleX509MngrVerifyCertAgainstCrl
							(xmlSecSimpleX509MngrPtr mngr, 
							 X509* cert);
int			xmlSecSimpleX509MngrVerifyCRL	(xmlSecSimpleX509MngrPtr mngr, 
							 X509_CRL *crl );


/**
 * xmlSecSimpleX509MngrCreate
 * @sign:
 *
 *
 *
 */
xmlSecSimpleX509MngrPtr	
xmlSecSimpleX509MngrCreate(void) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrCreate";
    xmlSecSimpleX509MngrPtr mngr;
    
    mngr = (xmlSecSimpleX509MngrPtr) xmlMalloc(sizeof(xmlSecSimpleX509Mngr));
    if(mngr == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlSecSimpleX509Mngr malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(mngr, 0, sizeof(xmlSecSimpleX509Mngr));

    mngr->xst = X509_STORE_new();
    if(mngr->xst == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create x509 store\n",
	    func);	
#endif 	    
	xmlSecSimpleX509MngrDestroy(mngr);
	return(NULL);
    }
    if(!X509_STORE_set_default_paths(mngr->xst)) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to set default paths\n",
	    func);	
#endif 	    
	xmlSecSimpleX509MngrDestroy(mngr);
	return(NULL);
    }
	
    mngr->knownCerts = sk_X509_new_null();
    if(mngr->knownCerts == NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create known certs store\n",
	    func);	
#endif 	    
	xmlSecSimpleX509MngrDestroy(mngr);
	return(NULL);
    }    

    mngr->crls = sk_X509_CRL_new_null();
    if(mngr->crls == NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create crls store\n",
	    func);	
#endif 	    
	xmlSecSimpleX509MngrDestroy(mngr);
	return(NULL);
    }    
    return(mngr);
}

/**
 * xmlSecSimpleX509MngrPtr
 *
 *
 *
 *
 */
void		
xmlSecSimpleX509MngrDestroy(xmlSecSimpleX509MngrPtr mngr) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrPtr";
    
    if(mngr == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: manager is null \n",
	    func);	
#endif 	    
	return;
    }
    if(mngr->xst != NULL) {
	X509_STORE_free(mngr->xst);
    }
    if(mngr->knownCerts != NULL) {
	sk_X509_pop_free(mngr->knownCerts, X509_free);
    }
    if(mngr->crls != NULL) {
	sk_X509_CRL_pop_free(mngr->crls, X509_CRL_free);
    }
    memset(mngr, 0, sizeof(xmlSecSimpleX509Mngr)); 
    xmlFree(mngr);
}

/**
 * xmlSecSimpleX509MngrAddTrustedCert
 *
 *
 *
 *
 */
int
xmlSecSimpleX509MngrAddTrustedCert(xmlSecSimpleX509MngrPtr mngr, const char *filename) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrAddTrustedCert";
    X509_LOOKUP *lookup = NULL; 
    int ret;
    
    if((mngr == NULL) || (mngr->xst == NULL) || (filename == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: manager, certs store or filename is null \n",
	    func);	
#endif 	    
	return(-1);
    }
    
    lookup = X509_STORE_add_lookup(mngr->xst, X509_LOOKUP_file());
    if(lookup == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: x509 file lookup creation failed\n",
	    func);	
#endif 	    
	return(-1);
    }

    ret = X509_LOOKUP_load_file(lookup, filename, X509_FILETYPE_PEM);
    if(ret != 1) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: x509 file \"%s\" load failed\n", 
	    func, filename);	
#endif 	    
	return(-1);
    }
    return(0);
}

/**
 * xmlSecSimpleX509MngrAddCert:
 *
 *
 */
int
xmlSecSimpleX509MngrAddCert(xmlSecSimpleX509MngrPtr mngr, const char *filename) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrAddCert";
    FILE *f;
    X509 *cert;
    
    if((mngr == NULL) || (mngr->knownCerts == NULL) || (filename == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: manager, knownCerts store or filename is null \n",
	    func);	
#endif 	    
	return(-1);
    }
    
    f = fopen(filename, "r");
    if(f == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to open file \"%s\" \n",
	    func, filename);	
#endif 	    
	return(-1);
    }
    
    cert = PEM_read_X509(f, NULL, NULL, NULL);
    if(cert == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read cert from file \"%s\" \n",
	    func, filename);	
#endif 	    
	fclose(f);
	return(-1);
    }    
    fclose(f);

    /* add cert to the list */
    sk_X509_push(mngr->knownCerts, cert);
    return(0);
}




/**
 * xmlSecSimpleX509MngrVerify
 *
 *
 *
 *
 */
int
xmlSecSimpleX509MngrVerify(xmlSecSimpleX509MngrPtr mngr, X509 *cert,
			 STACK_OF(X509) *untrusted) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrVerify";
    X509_STORE_CTX xsc; 
    int ret;  
    
    if((cert == NULL) || (mngr == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509_STORE or cert is null\n",
	    func);
#endif	    	
	return(-1);
    }

    /**
     * verify cert is not revoked
     */     
    ret = xmlSecSimpleX509MngrVerifyCertAgainstCrl(mngr, cert);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cert check against crls failed\n",
	    func);
#endif	    	
	return(-1);
    } else if(ret == 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cert is revoked\n",
	    func);
#endif	    	
	return(0);
    }
    
    /* walk thru all certs in untrusted and remove revoked certs */
    if(untrusted != NULL) {
	X509 *x;
	int i;
	
	for(i = 0; i < sk_num(untrusted); ) {	
	    x = sk_X509_value(untrusted, i);
	    if(x == cert) { 
		/* we checked cert above */
		++i; 
		continue;
	    }
	    ret = xmlSecSimpleX509MngrVerifyCertAgainstCrl(mngr, x);
	    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: cert check against crls failed\n",
		    func);
#endif	    	
		return(-1);	    
	    } else if(ret > 0) {
		++i;
	    } else {
		/* cert is revoked -- delete it from stack */
		sk_delete(untrusted, i);
		X509_free(x);
	    }
	}
    }
    
    X509_STORE_CTX_init (&xsc, mngr->xst, cert, untrusted);
    ret = X509_verify_cert(&xsc); 
    if(ret != 1) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cert verification failed (%d)\n",
	    func, xsc.error);
#endif	    	
    }
    X509_STORE_CTX_cleanup (&xsc);  
    return((ret == 1) ? 1 : 0);
}

/**
 *
 *
 *
 *
 */
int
xmlSecSimpleX509MngrVerifyCertAgainstCrl(xmlSecSimpleX509MngrPtr mngr, X509* cert) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrVerify";
    X509_NAME *issuer;
    X509_CRL *crl = NULL;
    X509_REVOKED *revoked;
    int i, n;
    int ret;  
    
    if((cert == NULL) || (mngr == NULL) || (mngr->crls == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509_STORE or cert is null\n",
	    func);
#endif	    	
	return(-1);
    }
    
    /*
     * Try to retrieve a CRL corresponding to the issuer of
     * the current certificate 
     */    
    n = sk_num(mngr->crls);
    for(i = 0; i < n; i++) {
	crl = sk_X509_CRL_value(mngr->crls, i);     
	issuer = X509_CRL_get_issuer(crl);
	if(xmlSecX509NamesCompare(X509_CRL_get_issuer(crl), issuer) == 0) { 
	    break;
	}
    }
    if((i >= n) || (crl == NULL)){
	/* no crls for this issuer */
	return(1);
    }

    /* 
     * Check date of CRL to make sure it's not expired 
     */
    ret = X509_cmp_current_time(X509_CRL_get_nextUpdate(crl));
    if (ret == 0) {
	/* crl expired */
	return(1);
    }
    
    /* 
     * Check if the current certificate is revoked by this CRL
     */
    n = sk_num(X509_CRL_get_REVOKED(crl));
    for (i = 0; i < n; i++) {
        revoked = (X509_REVOKED *)sk_value(X509_CRL_get_REVOKED(crl), i);
        if (ASN1_INTEGER_cmp(revoked->serialNumber, X509_get_serialNumber(cert)) == 0) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: certificate is revoked\n",
	        func);
#endif	    	
	    return(0);
        }
    }
    return(1);    
}

int
xmlSecSimpleX509MngrVerifyCRL(xmlSecSimpleX509MngrPtr mngr, X509_CRL *crl ) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrVerifyCRL";
    X509_STORE_CTX xsc; 
    X509_OBJECT xobj;
    EVP_PKEY *pkey;
    int ret;  
    
    if((crl == NULL) || (mngr == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509_STORE or crl is null\n",
	    func);
#endif	    	
	return(-1);
    }

    X509_STORE_CTX_init (&xsc, mngr->xst, NULL, NULL);
    ret = X509_STORE_get_by_subject(&xsc, X509_LU_X509, 
		    X509_CRL_get_issuer(crl), &xobj);
    if(ret <= 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Error getting CRL issuer certificate\n",
	    func);
#endif	    	
	return(-1);
    }
    pkey = X509_get_pubkey(xobj.data.x509);
    X509_OBJECT_free_contents(&xobj);
    if(pkey == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: Error getting CRL issuer public key\n",
	    func);
#endif	    	
	return(-1);
    }
    ret = X509_CRL_verify(crl, pkey);
    EVP_PKEY_free(pkey);    
    if(ret != 1) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: crl verification failed (%d)\n",
	    func, xsc.error);
#endif	    	
    }
    X509_STORE_CTX_cleanup (&xsc);  
    return((ret == 1) ? 1 : 0);
}

/**
 * xmlSecSimpleX509MngrAddCRL
 *
 *
 *
 *
 */
int
xmlSecSimpleX509MngrAddCRL(xmlSecSimpleX509MngrPtr mngr, X509_CRL *crl) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrAddCRL";
    int ret;
    
    if((crl == NULL) || (mngr == NULL) || (mngr->crls == NULL)){
#ifdef XMLSEC_DEBUG     
        xmlGenericError(xmlGenericErrorContext,
	    "%s: X509_STORE or crl is null\n",
	    func);
#endif	    	
	return(-1);
    }
    
    ret = xmlSecSimpleX509MngrVerifyCRL(mngr, crl);
    if(ret != 1) {
#ifdef XMLSEC_DEBUG     
        xmlGenericError(xmlGenericErrorContext,
	    "%s: crl not verified\n",
	    func);
#endif	    	
	return(-1);
    }    

    /** 
     * Nice function but according to the following message:
     *   http://groups.google.com/groups?hl=en&threadm=96uofi%2417gh%241%40FreeBSD.csie.NCTU.edu.tw&rnum=3&prev=/groups%3Fq%3Dopenssl%2BX509_STORE_add_crl%26hl%3Den%26selm%3D96uofi%252417gh%25241%2540FreeBSD.csie.NCTU.edu.tw%26rnum%3D3
     * this does not work (and I tried this with OpenSSL 0.9.6c with
     * the same result
     */
    X509_STORE_add_crl(mngr->xst, crl);
    /** 
     * so we have our own crls stack
     */
    sk_X509_CRL_push(mngr->crls, crl); 
    return(0);
}

/**
 * xmlSecSimpleX509MngrFind
 *
 *
 *
 *
 */
X509*		
xmlSecSimpleX509MngrFind(xmlSecSimpleX509MngrPtr mngr, xmlChar *subjectName,
			xmlChar *issuerName, xmlChar *issuerSerial,
			xmlChar *ski) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleX509MngrFind";
    X509 *cert = NULL;
    int i;
    
    if((mngr == NULL) || (mngr->knownCerts == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr is null\n",
	    func);	
#endif 	    
	return(NULL);    
    }

    /* todo: may be this is not the fastest way to search certs */
    if(subjectName != NULL) {
	X509_NAME *nm;
	X509_NAME *subj;

	nm = xmlSecX509NameRead(subjectName, xmlStrlen(subjectName));
	if(nm == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: subject name parsing failed\n",
		func);	
#endif 	    
	    return(NULL);    
	}

	for(i = 0; i < mngr->knownCerts->num; ++i) {
	    cert = ((X509**)(mngr->knownCerts->data))[i];
	    subj = X509_get_subject_name(cert);
	    if(xmlSecX509NamesCompare(nm, subj) == 0) {
		X509_NAME_free(nm);
		return(X509_dup(cert));
	    }	    
	}
	X509_NAME_free(nm);
    } else if((issuerName != NULL) && (issuerSerial != NULL)) {
	X509_NAME *nm;
	X509_NAME *issuer;
	BIGNUM *bn;
	ASN1_INTEGER *serial;

	nm = xmlSecX509NameRead(issuerName, xmlStrlen(issuerName));
	if(nm == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: issuer name parsing failed\n",
		func);	
#endif 	    
	    return(NULL);    
	}
		
	bn = BN_new();
	if(bn == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: BIGNUM creation failed\n",
		func);	
#endif 	    
	    X509_NAME_free(nm);
	    return(NULL);    
	}
	if(BN_dec2bn(&bn, (char*)issuerSerial) == 0) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: BIGNUM parsing failed\n",
		func);	
#endif 	    
	    BN_free(bn);
	    X509_NAME_free(nm);
	    return(NULL);    
	}
	
	serial = BN_to_ASN1_INTEGER(bn, NULL);
	if(serial == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: ASN1_INTEGER parsing failed\n",
		func);	
#endif 	    
	    BN_free(bn);
	    X509_NAME_free(nm);
	    return(NULL);    
	}
	BN_free(bn); 


	for(i = 0; i < mngr->knownCerts->num; ++i) {
	    cert = ((X509**)(mngr->knownCerts->data))[i];
	    if(ASN1_INTEGER_cmp(X509_get_serialNumber(cert), serial) != 0) {
		continue;
	    } 
	    issuer = X509_get_issuer_name(cert);
	    if(xmlSecX509NamesCompare(nm, issuer) == 0) {
		ASN1_INTEGER_free(serial);
		X509_NAME_free(nm);
		return(X509_dup(cert));
	    }	    
	}

        X509_NAME_free(nm);
	ASN1_INTEGER_free(serial);
    } else if(ski != NULL) {
	int len;
	int index;
	X509_EXTENSION *ext;
	ASN1_OCTET_STRING *keyId;
	
	/* our usual trick with base64 decode */
	len = xmlSecBase64Decode(ski, (unsigned char*)ski, xmlStrlen(ski));
	if(len < 0) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
	        "%s: failed to base64 decode ski\n",
		func);	
#endif 	    
	    return(NULL);    	
	}
	for(i = 0; i < mngr->knownCerts->num; ++i) {
	    cert = ((X509**)(mngr->knownCerts->data))[i];
	    index = X509_get_ext_by_NID(cert, NID_subject_key_identifier, -1); 
	    if((index >= 0)  && (ext = X509_get_ext(cert, index))) {
		keyId = X509V3_EXT_d2i(ext);
		if((keyId != NULL) && (keyId->length == len) && 
				    (memcmp(keyId->data, ski, len) == 0)) {
		    M_ASN1_OCTET_STRING_free(keyId);
		    return(X509_dup(cert));
		}
		M_ASN1_OCTET_STRING_free(keyId);
	    }
	}	
    }

    return(NULL);
}

/**
 * xmlSecX509NameRead
 *
 *
 */       
static X509_NAME *
xmlSecX509NameRead(unsigned char *str, int len) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509NameRead";
    unsigned char name[256];
    unsigned char value[256];
    int nameLen, valueLen;
    X509_NAME *nm;
    int type = MBSTRING_ASC;
    
    nm = X509_NAME_new();
    if(nm == NULL) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create X509_NAME\n",
	    func);	
#endif 		    
	return(NULL);
    }
    
    while(len > 0) {
	/* skip spaces after comma or semicolon */
	while((len > 0) && isspace(*str)) {
	    ++str; --len;
	}

	nameLen = xmlSecX509NameStringRead(&str, &len, name, sizeof(name), '=', 0);	
	if(nameLen < 0) {
#ifdef XMLSEC_DEBUG
	    xmlGenericError(xmlGenericErrorContext,
		"%s: name read failed\n",
		func);	
#endif 		    
	    X509_NAME_free(nm);
	    return(NULL);
	}
	name[nameLen] = '\0';
	if(len > 0) {
	    ++str; --len;
	    if((*str) == '\"') {
		valueLen = xmlSecX509NameStringRead(&str, &len, 
					value, sizeof(value), '"', 1);	
		if(valueLen < 0) {
#ifdef XMLSEC_DEBUG
		    xmlGenericError(xmlGenericErrorContext,
			"%s: failed to read quoted value\n",
			func);	
#endif 		    
		    X509_NAME_free(nm);
		    return(NULL);
    		}
		/* skip spaces before comma or semicolon */
		while((len > 0) && isspace(*str)) {
		    ++str; --len;
		}
		if((len > 0) && ((*str) != ',')) {
#ifdef XMLSEC_DEBUG
		    xmlGenericError(xmlGenericErrorContext,
			"%s: comma is expected\n",
			func);	
#endif 		    
		    X509_NAME_free(nm);
		    return(NULL);
		}
		if(len > 0) {
		    ++str; --len;
		}
		type = MBSTRING_ASC;
	    } else if((*str) == '#') {
		    /* TODO: */
#ifdef XMLSEC_DEBUG
		    xmlGenericError(xmlGenericErrorContext,
			"%s: reading octect values is not implemented yet\n",
			func);	
#endif 		    
    	        X509_NAME_free(nm);
		return(NULL);
	    } else {
		valueLen = xmlSecX509NameStringRead(&str, &len, 
					value, sizeof(value), ',', 1);	
		if(valueLen < 0) {
#ifdef XMLSEC_DEBUG
		    xmlGenericError(xmlGenericErrorContext,
			"%s: failed to read string value\n",
			func);	
#endif 		    
    	    	    X509_NAME_free(nm);
		    return(NULL);
    		}
		type = MBSTRING_ASC;
	    } 			
	} else {
	    valueLen = 0;
	}
	value[valueLen] = '\0';
	if(len > 0) {
	    ++str; --len;
	}	
	X509_NAME_add_entry_by_txt(nm, (char*)name, type, value, valueLen, -1, 0);
    }
    
    return(nm);
}


/** 
 * we have to sort X509_NAME entries to get correct results.
 * This is ugly but OpenSSL does not support it
 */
static int		
xmlSecX509NamesCompare(X509_NAME *a, X509_NAME *b) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509NamesCompare";
    X509_NAME *a1 = NULL;
    X509_NAME *b1 = NULL;
    int ret;
    
    
    if(a != NULL) {
	a1 = X509_NAME_dup(a);
    }
    if(a1 == NULL) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
		"%s: X509_NAME_dup(a) failed\n",
		func);	
#endif 		    
        return(-1);
    }
    if(b != NULL) {
	b1 = X509_NAME_dup(b);
    }
    if(b1 == NULL) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
		"%s: X509_NAME_dup(b) failed\n",
		func);	
#endif 		
	X509_NAME_free(a1);    
        return(1);
    }
        
    /* sort both */
    sk_X509_NAME_ENTRY_set_cmp_func(a1->entries, xmlSecX509_NAME_ENTRY_cmp);
    sk_X509_NAME_ENTRY_sort(a1->entries);
    sk_X509_NAME_ENTRY_set_cmp_func(b1->entries, xmlSecX509_NAME_ENTRY_cmp);
    sk_X509_NAME_ENTRY_sort(b1->entries);

    /* actually compare */
    ret = xmlSecX509_NAME_cmp(a1, b1);
    
    /* cleanup */
    X509_NAME_free(a1);
    X509_NAME_free(b1);
    return(ret);
}
			

/**
 * xmlSecX509_NAME_ENTRY_cmp
 *
 *
 *
 */
static int 
xmlSecX509_NAME_ENTRY_cmp(const X509_NAME_ENTRY **a, const X509_NAME_ENTRY **b) {
    return(OBJ_cmp((*a)->object, (*b)->object));
}

/**
 * xmlSecX509NameStringRead
 *
 *
 *
 */
static int 
xmlSecX509NameStringRead(unsigned char **str, int *strLen, 
			unsigned char *res, int resLen,
			unsigned char delim, int ingoreTrailingSpaces) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecX509NameStringRead";
    unsigned char *p, *q, *nonSpace; 
    
    p = (*str);
    nonSpace = q = res;
    while(((p - (*str)) < (*strLen)) && ((*p) != delim) && ((q - res) < resLen)) { 
	if((*p) != '\\') {
	    if(ingoreTrailingSpaces && !isspace(*p)) nonSpace = q;	
	    *(q++) = *(p++);
	} else {
	    ++p;
	    nonSpace = q;    
	    if(xmlSecIsHex((*p))) {
		if((p - (*str) + 1) >= (*strLen)) {
#ifdef XMLSEC_DEBUG
		    xmlGenericError(xmlGenericErrorContext,
			"%s: two hex digits expected\n",
			func);	
#endif 		    
	    	    return(-1);
		}
		*(q++) = xmlSecGetHex(p[0]) * 16 + xmlSecGetHex(p[1]);
		p += 2;
	    } else {
		if(((++p) - (*str)) >= (*strLen)) {
#ifdef XMLSEC_DEBUG
		    xmlGenericError(xmlGenericErrorContext,
			"%s: escaped symbol missed\n",
			func);	
#endif 		    
		    return(-1);
		}
		*(q++) = *(p++); 
	    }
	}	    
    }
    if(((p - (*str)) < (*strLen)) && ((*p) != delim)) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: buffer is too small\n",
	    func);	
#endif 		    
	return(-1);
    }
    (*strLen) -= (p - (*str));
    (*str) = p;
    return((ingoreTrailingSpaces) ? nonSpace - res + 1 : q - res);
}

static
int xmlSecX509_NAME_cmp(const X509_NAME *a, const X509_NAME *b)
	{
	int i,j;
	X509_NAME_ENTRY *na,*nb;

	if (sk_X509_NAME_ENTRY_num(a->entries)
	    != sk_X509_NAME_ENTRY_num(b->entries))
		return sk_X509_NAME_ENTRY_num(a->entries)
		  -sk_X509_NAME_ENTRY_num(b->entries);
	for (i=sk_X509_NAME_ENTRY_num(a->entries)-1; i>=0; i--)
		{
		na=sk_X509_NAME_ENTRY_value(a->entries,i);
		nb=sk_X509_NAME_ENTRY_value(b->entries,i);
		j=na->value->length-nb->value->length;
		if (j) return(j);
		j=memcmp(na->value->data,nb->value->data,
			na->value->length);
		if (j) return(j);
		}

	/* We will check the object types after checking the values
	 * since the values will more often be different than the object
	 * types. */
	for (i=sk_X509_NAME_ENTRY_num(a->entries)-1; i>=0; i--)
		{
		na=sk_X509_NAME_ENTRY_value(a->entries,i);
		nb=sk_X509_NAME_ENTRY_value(b->entries,i);
		j=OBJ_cmp(na->object,nb->object);
		if (j) return(j);
		}
	return(0);
	}


#endif /* XMLSEC_NO_X509 */


