/** 
 * XMLSec library
 *
 * "XML Encryption" implementation
 *  http://www.w3.org/TR/xmlenc-core
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#ifndef XMLSEC_NO_XMLENC
 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h> 

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysInternal.h>
#include <xmlsec/transforms.h>
#include <xmlsec/transformsInternal.h>
#include <xmlsec/xmlenc.h>
#include <xmlsec/base64.h>
#include <xmlsec/membuf.h>


const xmlChar xmlEncTypeElement[] = "http://www.w3.org/2001/04/xmlenc#Element";
const xmlChar xmlEncTypeContent[] = "http://www.w3.org/2001/04/xmlenc#Content";

typedef struct _xmlEncState {
    xmlEncCtxPtr		ctx;
    xmlSecBinTransformPtr 	first;
    xmlSecBinTransformPtr	last;
    xmlNodePtr			cipherDataNode;
    int				encrypt;
    xmlChar 			*type;
    
} xmlEncState, *xmlEncStatePtr;

/** 
 * XML Enc state
 */
static xmlEncStatePtr		xmlEncStateCreate	(xmlEncCtxPtr ctx, 
							 int encrypt);
static void			xmlEncStateDestroy	(xmlEncStatePtr state);
static int			xmlEncStateAddFirstTransform(xmlEncStatePtr state,
							 xmlSecTransformPtr transform);
static int			xmlEncStateAddTransform	(xmlEncStatePtr state,
							 xmlSecTransformPtr transform);

#define xmlEncStateGetKeysContext(state) \
    ((( ( state ) != NULL ) && ( ( state )->ctx != NULL ) ) ? \
      (( ( state )->ctx->keysReadContext)) : NULL)


static int			xmlEncEncryptedDataRead	(xmlEncStatePtr state, 
							 xmlNodePtr encData,
							 xmlEncResultPtr result);
static int			xmlEncChipherDataNodeRead(xmlEncCtxPtr ctx, 
							 xmlEncStatePtr state, 							 
							 xmlNodePtr node,
							 xmlEncResultPtr result);
static int 			xmlEncCipherDataNodeWrite(xmlNodePtr cipherDataNode,
							 const unsigned char *buf,
							 size_t size);
static int			xmlEncChipherValueNodeRead(xmlEncCtxPtr ctx, 
							 xmlEncStatePtr state, 
							 xmlNodePtr node,
							 xmlEncResultPtr result);
static int			xmlEncChipherRefNodeRead(xmlEncCtxPtr ctx, 
							 xmlEncStatePtr state, 
							 xmlNodePtr node,
							 xmlEncResultPtr result);

static int			xmlEncReplaceNode	(xmlNodePtr node,
							 xmlNodePtr newNode);
static int			xmlEncReplaceContent	(xmlNodePtr node,
							 xmlNodePtr newNode);
static int			xmlEncReplaceNodeBuffer	(xmlNodePtr node,
							 const unsigned char *buffer, 
							 size_t size);
static int			xmlEncEncryptInit	(xmlEncCtxPtr ctx, 
							 xmlNodePtr encData,
							 xmlEncStatePtr *state, 
							 xmlEncResultPtr *result);
							 
static int			xmlEncEncryptFinalize	(xmlEncCtxPtr ctx, 
							 xmlNodePtr node,
							 xmlEncStatePtr state, 
							 xmlEncResultPtr result);


/**
 * XML Encrypiton context methods
 */
/**
 * xmlEncCtxCreate:
 *
 *
 *
 */
xmlEncCtxPtr		
xmlEncCtxCreate(const xmlSecKeysReadContext *keysReadContext) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncCtxCreate";
    xmlEncCtxPtr ctx;
    
    /*
     * Allocate a new xmlEncCtx and fill the fields.
     */
    ctx = (xmlEncCtxPtr) xmlMalloc(sizeof(xmlEncCtx));
    if(ctx == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlEncCtx malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(ctx, 0, sizeof(xmlEncCtx));
    
    ctx->keysReadContext = keysReadContext;
    return(ctx);
}

/**
 * xmlEncCtxDestroy
 *
 *
 *
 */
void
xmlEncCtxDestroy(xmlEncCtxPtr ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncCtxDestroy";
    if(ctx == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n", 
	    func);	
#endif
	return;	    
    }
    
    memset(ctx, 0, sizeof(xmlEncCtx));
    xmlFree(ctx);
}


/** 
 * Encryption Template
 */
/** 
 * xmlEncTmplCreate:
 * @id: the Id attribute of EncryptedData node (optional)
 * @type: the Type attribute of EncryptedData node (optional)
 * @mimeType: the MimeType attribute of EncryptedData node (optional)
 * @encoding: the Encoding attribute of EncryptedData node (optional)
 *
 * Creates new encryption template. 
 *
 * Returns the new template or NULL if an error occurs
 */
xmlEncTmplPtr		
xmlEncTmplCreate(const xmlChar *id, const xmlChar *type,
		 const xmlChar *mimeType, const xmlChar *encoding) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplCreate";
    xmlDocPtr doc;
    xmlNodePtr encData;
    xmlNodePtr cipherData;
    
    doc = xmlNewDoc(NULL);
    if(doc == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create new doc\n", 
	    func);	
#endif
	return(NULL);	        
    }
    encData = xmlNewDocNode(doc, NULL, BAD_CAST "EncryptedData", NULL);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create new node\n", 
	    func);	
#endif
	xmlFreeDoc(doc);
	return(NULL);	        
    }
    xmlDocSetRootElement(doc, encData);
    
    if(xmlNewNs(encData, xmlSecEncNs, NULL) == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add namespace\n", 
	    func);	
#endif
	xmlFreeDoc(doc);
	return(NULL);	        	
    }
    
    if(id != NULL) {
	xmlSetProp(encData, BAD_CAST "Id", id);
    }
    if(type != NULL) {
	xmlSetProp(encData, BAD_CAST "Type", type);
    }
    if(mimeType != NULL) {
	xmlSetProp(encData, BAD_CAST "MimeType", mimeType);
    }
    if(encoding != NULL) {
	xmlSetProp(encData, BAD_CAST "Encoding", encoding);
    }
    
    cipherData = xmlSecAddChild(encData,  BAD_CAST "CipherData", xmlSecEncNs);
    if(cipherData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add CipherData\n", 
	    func);	
#endif
	xmlFreeDoc(doc);
	return(NULL);	        	
    }
    
    return((xmlEncTmplPtr)encData);
}

/** 
 * xmlEncTmplDestroy
 *
 *
 *
 */
void
xmlEncTmplDestroy(xmlEncTmplPtr tmpl) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplDestroy";
    xmlNodePtr encData;
        
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return;	        
    }

    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
    	    func);	
#endif
	return;	
    }

    if(encData == xmlDocGetRootElement(encData->doc)) {
        xmlFreeDoc(encData->doc);
    } else {
	xmlUnlinkNode(encData);
	xmlFreeNode(encData);
    }
}

/** 
 * xmlEncTmplAddEncMethod
 *
 *
 *
 */
xmlNodePtr
xmlEncTmplAddEncMethod(xmlEncTmplPtr tmpl, xmlSecTransformId encMethodId) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddEncMethod";
    xmlNodePtr encData;
    xmlNodePtr encMethod;
    xmlNodePtr tmp;
    int ret;
    
    if((tmpl == NULL) || (encMethodId == xmlSecTransformUnknown)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null or transform unknown\n", 
	    func);	
#endif
	return(NULL);	        
    }
    
    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    encMethod = xmlSecFindChild(encData, BAD_CAST "EncryptionMethod", xmlSecEncNs);
    if(encMethod != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the EncryptionMethod node is already there\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    tmp = xmlSecGetNextElementNode(encData->children);
    if(tmp == NULL) {
	encMethod = xmlSecAddChild(encData,  BAD_CAST "EncryptionMethod", xmlSecEncNs);
    } else {
	encMethod = xmlSecAddPrevSibling(tmp,  BAD_CAST "EncryptionMethod", xmlSecEncNs);
    }    
    if(encMethod == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create EncryptionMethod node\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    ret = xmlSecTransformNodeWrite(encMethod, encMethodId);
    if(ret < 0){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: enc method write failed\n", 
	    func);	
#endif
	xmlUnlinkNode(encMethod);
	xmlFreeNode(encMethod);
	return(NULL);	
    }
    return(encMethod);
}

/** 
 * xmlEncTmplAddKeyInfo
 *
 *
 *
 */
xmlNodePtr
xmlEncTmplAddKeyInfo(xmlEncTmplPtr tmpl) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddKeyInfo";
    xmlNodePtr keyInfo;
    xmlNodePtr encData;
    xmlNodePtr prev;
    xmlNodePtr tmp;
        
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return(NULL);	        
    }
    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }

    keyInfo = xmlSecFindChild(encData, BAD_CAST "KeyInfo", xmlSecDSigNs);
    if(keyInfo != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the KeyInfo node is already there\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    prev = xmlSecFindChild(encData, BAD_CAST "EncryptionMethod", xmlSecEncNs);
    tmp = xmlSecGetNextElementNode(encData->children);
    if(prev != NULL) {
	keyInfo = xmlSecAddNextSibling(prev, BAD_CAST "KeyInfo", xmlSecDSigNs);
    } else if(tmp == NULL) {
	keyInfo = xmlSecAddChild(encData, BAD_CAST "KeyInfo", xmlSecDSigNs);
    } else {
	keyInfo = xmlSecAddPrevSibling(tmp, BAD_CAST "KeyInfo", xmlSecDSigNs);
    }
    if(keyInfo == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create keyInfo node\n", 
	    func);	
#endif
	return(NULL);	
    }
    return(keyInfo);
}

/** 
 * xmlEncTmplAddEncProperties
 *
 *
 *
 */
xmlNodePtr
xmlEncTmplAddEncProperties(xmlEncTmplPtr tmpl, const xmlChar *id) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddEncProperties";
    xmlNodePtr encProps;
    xmlNodePtr encData;
        
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return(NULL);	        
    }

    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }

    encProps = xmlSecFindChild(encData, BAD_CAST "EncryptionProperties", xmlSecEncNs);
    if(encProps != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the EncryptionProperties node is already there\n", 
	    func);	
#endif
	return(NULL);	
    }

    encProps = xmlSecAddChild(encData, BAD_CAST "EncryptionProperties", xmlSecEncNs);
    if(encProps == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create EncryptionProperties node\n", 
	    func);	
#endif
	return(NULL);	
    }
    if(id != NULL) {
	xmlSetProp(encProps, BAD_CAST "Id", id);
    }
    
    return(encProps);
}

/** 
 * xmlEncTmplAddEncProperty
 *
 *
 *
 */
xmlNodePtr	
xmlEncTmplAddEncProperty(xmlEncTmplPtr tmpl, const xmlChar *id,  const xmlChar *target) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddEncProperty";
    xmlNodePtr encProp;
    xmlNodePtr encProps;
    xmlNodePtr encData;
        
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return(NULL);	        
    }

    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }

    encProps = xmlSecFindChild(encData, BAD_CAST "EncryptionProperties", xmlSecEncNs);
    if(encProps == NULL) {
	encProps = xmlEncTmplAddEncProperties(tmpl, NULL);
	if(encProps == NULL) { 
#ifdef XMLSEC_DEBUG
            xmlGenericError(xmlGenericErrorContext,
		"%s: the EncryptionProperties node creatin failed\n", 
	        func);	
#endif
	    return(NULL);	
	}
    }

    encProp = xmlSecAddChild(encProps, BAD_CAST "EncryptionProperty", xmlSecEncNs);
    if(encProp == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create EncryptionProperty node\n", 
	    func);	
#endif
	return(NULL);	
    }
    if(id != NULL) {
	xmlSetProp(encProp, BAD_CAST "Id", id);
    }
    if(target != NULL) {
	xmlSetProp(encProp, BAD_CAST "Target", target);
    }
    
    return(encProp);
}

/** 
 * xmlEncTmplAddCipherValue
 *
 *
 *
 */
xmlNodePtr
xmlEncTmplAddCipherValue(xmlEncTmplPtr tmpl) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddCipherValue";
    xmlNodePtr encData;
    xmlNodePtr cipherData;
    xmlNodePtr cipherValue;
    xmlNodePtr tmp;
        
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return(NULL);	        
    }

    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }

    cipherData = xmlSecFindChild(encData, BAD_CAST "CipherData", xmlSecEncNs);
    if(cipherData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherData node is not found\n", 
	    func);	
#endif
	return(NULL);	
    }

    tmp = xmlSecFindChild(cipherData, BAD_CAST "CipherValue", xmlSecEncNs);
    if(tmp != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherValue node is already present\n", 
	    func);	
#endif
	return(NULL);	
    }

    tmp = xmlSecFindChild(cipherData, BAD_CAST "CipherReference", xmlSecEncNs);
    if(tmp != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherReference node is already present\n", 
	    func);	
#endif
	return(NULL);	
    }

    cipherValue = xmlSecAddChild(cipherData, BAD_CAST "CipherValue", xmlSecEncNs);
    if(cipherValue == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create CipherValue node\n", 
	    func);	
#endif
	return(NULL);	
    }    
        
    return(cipherValue);
}

/** 
 * xmlEncTmplAddCipherRef
 *
 *
 *
 */
xmlNodePtr
xmlEncTmplAddCipherRef(xmlEncTmplPtr tmpl, const xmlChar *uri) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddCipherRef";
    xmlNodePtr cipherRef;
    xmlNodePtr cipherData;    
    xmlNodePtr encData;
    xmlNodePtr tmp;
    
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return(NULL);	        
    }
    encData = xmlEncTmplGetEncData(tmpl);
    if(encData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }

    cipherData = xmlSecFindChild(encData, BAD_CAST "CipherData", xmlSecEncNs);
    if(cipherData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherData node is not found\n", 
	    func);	
#endif
	return(NULL);	
    }

    tmp = xmlSecFindChild(cipherData, BAD_CAST "CipherValue", xmlSecEncNs);
    if(tmp != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherValue node is already present\n", 
	    func);	
#endif
	return(NULL);	
    }

    tmp = xmlSecFindChild(cipherData, BAD_CAST "CipherReference", xmlSecEncNs);
    if(tmp != NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherReference node is already present\n", 
	    func);	
#endif
	return(NULL);	
    }

    cipherRef = xmlSecAddChild(cipherData, BAD_CAST "CipherReference", xmlSecEncNs);
    if(cipherRef == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create CipherValue node\n", 
	    func);	
#endif
	return(NULL);	
    }    
    
    if(uri != NULL) {
	xmlSetProp(cipherRef, BAD_CAST "URI", uri);
    }
    
    return(cipherRef);
}

/** 
 * xmlEncTmplAddCipherRefTransform
 *
 *
 *
 */
xmlNodePtr
xmlEncTmplAddCipherRefTransform	(xmlEncTmplPtr tmpl, xmlSecTransformId transform) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncTmplAddCipherRefTransform";
    xmlNodePtr encData;
    xmlNodePtr cipherData;
    xmlNodePtr cipherRef;    
    xmlNodePtr transforms;
    xmlNodePtr cipherRefTransform;
    int ret;
    
    if((tmpl == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: template is null\n", 
	    func);	
#endif
	return(NULL);	        
    }

    encData = xmlEncTmplGetEncData(tmpl);
    if(encData== NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get EncData node\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    cipherData = xmlSecFindChild(encData, BAD_CAST "CipherData", xmlSecEncNs);
    if(cipherData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherData node is not found\n", 
	    func);	
#endif
	return(NULL);	
    }

    cipherRef = xmlSecFindChild(cipherData, BAD_CAST "CipherReference", xmlSecEncNs);
    if(cipherRef == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherReference node is not found\n", 
	    func);	
#endif
	return(NULL);	
    }

    transforms = xmlSecFindChild(cipherRef, BAD_CAST "Transforms", xmlSecEncNs);
    if(transforms == NULL) {
	transforms = xmlSecAddChild(cipherRef, BAD_CAST "Transforms", xmlSecEncNs);
	if(transforms == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create Transforms node\n", 
		func);	
#endif
	    return(NULL);	
	}
    }
    
    cipherRefTransform = xmlSecAddChild(transforms,  BAD_CAST "Transform", xmlSecDSigNs);
    if(cipherRefTransform == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create Transform node\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    ret = xmlSecTransformNodeWrite(cipherRefTransform, transform);
    if(ret < 0){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: enc method write failed\n", 
	    func);	
#endif
	return(NULL);	
    }
    
    return(cipherRefTransform);
}


/**
 * xmlKeyInfoAddEncryptedKey:
 * @keyInfoNode:
 * @id:
 * @type:
 * @recipient:
 *
 */
xmlEncTmplPtr		
xmlKeyInfoAddEncryptedKey(xmlNodePtr keyInfoNode, const xmlChar *id,
			 const xmlChar *type, const xmlChar *recipient) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlKeyInfoAddEncryptedKey";
    xmlNodePtr encKey;
    xmlNodePtr cipherData;
    
    if((keyInfoNode == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: keyInfoNode is null\n",
	    func);	
#endif
	return(NULL);	
    }

    /* we allow multiple encrypted key elements */
    encKey = xmlSecAddChild(keyInfoNode, BAD_CAST "EncryptedKey", xmlSecEncNs); 
    if(encKey == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create EncryptedKey node\n",
	    func);	
#endif
	return(NULL);	
    }
    
    
    if(id != NULL) {
	xmlSetProp(encKey, BAD_CAST "Id", id);
    }
    if(type != NULL) {
	xmlSetProp(encKey, BAD_CAST "Type", type);
    }
    if(recipient != NULL) {
	xmlSetProp(encKey, BAD_CAST "Recipient", recipient);
    }

    cipherData = xmlSecAddChild(encKey,  BAD_CAST "CipherData", xmlSecEncNs);
    if(cipherData == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add CipherData\n", 
	    func);	
#endif
	xmlUnlinkNode(encKey);
	xmlFreeNode(encKey);
	return(NULL);	        	
    }    
    return((xmlEncTmplPtr)encKey);    
}


/**
 * Encryption
 */
/**
 * xmlEncEncryptInit
 *
 *
 *
 *
 */
static int
xmlEncEncryptInit(xmlEncCtxPtr ctx, xmlNodePtr encData,
		xmlEncStatePtr *state, xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptInit";
    int ret;
    
    if((ctx == NULL) || (encData == NULL) || (state == NULL) || (result == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, encData node, state or result is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    /* first of all, create result and encryption state objects */
    (*state) = xmlEncStateCreate(ctx, 1);
    if((*state) == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create encryption state\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    (*result) = xmlEncResultCreate(ctx, 1, encData);
    if((*result) == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create result object\n", 
	    func);	
#endif
	return(-1);	    		
    }
    
    /*  read and update the Encryption Method and KeyInfo */
    ret = xmlEncEncryptedDataRead((*state), encData, (*result));
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read EncryptedData\n", 
	    func);	
#endif
	return(-1);	    	
    }
    
    return(0);    
}

/**
 * xmlEncEncryptFinalize
 *
 *
 *
 */
static int
xmlEncEncryptFinalize(xmlEncCtxPtr ctx, xmlNodePtr encData,
		    xmlEncStatePtr state, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptFinalize";
    int ret;
    
    if((ctx == NULL) || (encData == NULL) || (state == NULL) || (result == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, encData, state or result is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    /* update template */
    result->buffer = xmlSecMemBufTransformGetBuffer((xmlSecTransformPtr)state->last, 1);
    if(result->buffer == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get memory buffer\n", 
	    func);	
#endif
	xmlEncResultDestroy(result);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    
    ret = xmlEncCipherDataNodeWrite(state->cipherDataNode, 
				xmlBufferContent(result->buffer), 
				xmlBufferLength(result->buffer));
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write CipherData node\n", 
	    func);	
#endif
	xmlEncResultDestroy(result);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    return(0);
}

/**
 * xmlEncEncryptMemory
 *
 *
 *
 */
int
xmlEncEncryptMemory(xmlEncCtxPtr ctx, xmlEncTmplPtr tmpl,
		    const unsigned char *buf, size_t size,
		    xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptMemory";
    xmlEncStatePtr state = NULL;
    xmlEncResultPtr res = NULL;
    int ret;
    
    if((ctx == NULL) || (tmpl == NULL) || (buf == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, tmpl or buff is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    /**
     * Init everything 
     */    
    ret = xmlEncEncryptInit(ctx, xmlEncTmplGetEncData(tmpl), &state, &res);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: initialization failed\n", 
	    func);	
#endif
	if(state != NULL) xmlEncStateDestroy(state);
	if(res != NULL) xmlEncResultDestroy(res); 
	return(-1);	    
    }
    
    
    /* encrypt the data */
    ret = xmlSecBinTransformWrite((xmlSecTransformPtr)state->first, buf, size);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to encrypt the buffer\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    ret = xmlSecBinTransformFlush((xmlSecTransformPtr)state->first);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize encryption\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }

    ret = xmlEncEncryptFinalize(ctx, xmlEncTmplGetEncData(tmpl), state, res); 			
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: finalization failed\n", 
	    func);	
#endif
	xmlEncStateDestroy(state);
	xmlEncResultDestroy(res); 
	return(-1);	    
    }
    
    /* cleanup */
    if(result != NULL) {
	(*result) = res;
    } else {
	xmlEncResultDestroy(res);
    }
    xmlEncStateDestroy(state);    
    return(0);
}

/**
 * xmlEncEncryptUri
 *
 *
 *
 */
int
xmlEncEncryptUri(xmlEncCtxPtr ctx, xmlEncTmplPtr tmpl, const char *uri,
		 xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptUri";

    if((ctx == NULL) || (tmpl == NULL) || (uri == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, tmpl or uri is null\n", 
	    func);	
#endif
	return(-1);	    
    }

#ifdef XMLSEC_DEBUG
    xmlGenericError(xmlGenericErrorContext,
	"%s: not implemented\n", 
	func);	
#endif
    return(-1);	    

    /* TODO: */
    return(0);
}

/**
 * xmlEncEncryptXmlNode
 *
 *
 *
 */
int
xmlEncEncryptXmlNode(xmlEncCtxPtr ctx, xmlEncTmplPtr tmpl, xmlNodePtr src,
		 xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptXmlNode";
    xmlEncStatePtr state = NULL;
    xmlEncResultPtr res = NULL;
    xmlBufferPtr buffer = NULL;
    int ret;

    if((ctx == NULL) || (tmpl == NULL) || (src == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context, tmpl or src is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    /**
     * Init everything 
     */    
    ret = xmlEncEncryptInit(ctx, xmlEncTmplGetEncData(tmpl), &state, &res);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: initialization failed\n", 
	    func);	
#endif
	if(state != NULL) xmlEncStateDestroy(state);
	if(res != NULL) xmlEncResultDestroy(res); 
	return(-1);	    
    }
    
    buffer = xmlBufferCreate();
    if(buffer == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: output buffer creation failed\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    
    /** 
     * read the stuff we need to encrypt into the buffer 
     */
    if(ctx->ignoreType || (res->type == NULL)) {
	/* get the content of the node */
	xmlNodeDump(buffer, src->doc, src, 0, 0);
    } else if(xmlStrEqual(res->type, xmlEncTypeElement)) {
	/* get the content of the node */
	xmlNodeDump(buffer, src->doc, src, 0, 0);
    } else if(xmlStrEqual(res->type, xmlEncTypeContent)) {
	xmlNodePtr ptr;
	/* get the content of the nodes childs */
	ptr = src->children;
	while(ptr != NULL) {
	    xmlNodeDump(buffer, ptr->doc, ptr, 0, 0); 
	    ptr = ptr->next;
	}
    } else {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: the type \"%s\" is unknown\n", 
	    func, res->type);	
#endif
	xmlBufferFree(buffer);
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
        
    /* encrypt the data */
    ret = xmlSecBinTransformWrite((xmlSecTransformPtr)state->first, 
				  xmlBufferContent(buffer),
				  xmlBufferLength(buffer));
    xmlBufferFree(buffer); buffer = NULL;
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to encrypt the buffer\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }
    ret = xmlSecBinTransformFlush((xmlSecTransformPtr)state->first);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize encryption\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    
	return(-1);	    	
    }

    ret = xmlEncEncryptFinalize(ctx, xmlEncTmplGetEncData(tmpl), state, res); 			
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: finalization failed\n", 
	    func);	
#endif
	xmlEncStateDestroy(state);
	xmlEncResultDestroy(res); 
	return(-1);	    
    }
    
    if(!ctx->ignoreType && (res->type != NULL)) {
	if(xmlStrEqual(res->type, xmlEncTypeElement)) {
	    ret = xmlEncReplaceNode(src, xmlEncTmplGetEncData(tmpl));
	    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: node replacement failed\n", 
		    func);	
#endif
		xmlEncStateDestroy(state);
		xmlEncResultDestroy(res); 
		return(-1);
	    }
	    res->replaced = 1;			       
	} else if(xmlStrEqual(res->type, xmlEncTypeContent)) {
	    ret = xmlEncReplaceContent(src, xmlEncTmplGetEncData(tmpl));
	    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: content replacement failed\n", 
		    func);	
#endif
		xmlEncStateDestroy(state);
		xmlEncResultDestroy(res); 
		return(-1);
	    }
	    res->replaced = 1;			       
	} else {
	    /* do nothing: type is unknown */
	}
    }
    
    /* cleanup */
    if(result != NULL) {
	(*result) = res;
    } else {
	xmlEncResultDestroy(res);
    }
    xmlEncStateDestroy(state);    
    return(0);
}

/**
 * Decryption
 */
 
/**
 * xmlEncDecrypt
 *
 *
 *
 *
 */
int
xmlEncDecrypt(xmlEncCtxPtr ctx, xmlNodePtr encDataNode, xmlEncResultPtr *result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncDecrypt";
    xmlEncStatePtr state;
    xmlEncResultPtr res;
    int ret;
    
    if((ctx == NULL) || (encDataNode == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context or encDataNode is null\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    /* first of all, create result and encryption state objects */
    state = xmlEncStateCreate(ctx, 0);
    if(state == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create encryption state\n", 
	    func);	
#endif
	return(-1);	    
    }
    
    res = xmlEncResultCreate(ctx, 0, encDataNode);
    if(res == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create result object\n", 
	    func);	
#endif
	xmlEncStateDestroy(state);    
	return(-1);	    		
    }
    
    /*  read and update the Encryption Method and KeyInfo */
    ret = xmlEncEncryptedDataRead(state, encDataNode, res);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to read EncryptedData\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    	
	return(-1);	    	
    }
    
    if(state->cipherDataNode == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: CipherData node is not found\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    	
	return(-1);	    	
    }
    
    ret = xmlEncChipherDataNodeRead(ctx, state, state->cipherDataNode, res);
    if((ret < 0) || (res->buffer == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get CipherData node content\n", 
	    func);	
#endif
	xmlEncResultDestroy(res);
	xmlEncStateDestroy(state);    	
	return(-1);	    	
    }
    
    if(!ctx->ignoreType && (res->type != NULL)) {
	if(xmlStrEqual(res->type, xmlEncTypeElement)) {
	    ret = xmlEncReplaceNodeBuffer(encDataNode, 
				    xmlBufferContent(res->buffer),  
				    xmlBufferLength(res->buffer));
	    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to replace node\n", 
		    func);	
#endif
		xmlEncResultDestroy(res);
		xmlEncStateDestroy(state);    
		return(-1);	    	
	    }
	    res->replaced = 1;			       
	} else if(xmlStrEqual(res->type, xmlEncTypeContent)) {
	    /* replace the node with the buffer */
	    ret = xmlEncReplaceNodeBuffer(encDataNode, 
				       xmlBufferContent(res->buffer),
				       xmlBufferLength(res->buffer));
	    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to replace content\n", 
		    func);	
#endif
		xmlEncResultDestroy(res);
		xmlEncStateDestroy(state);    
		return(-1);	    	
	    }	
	    res->replaced = 1;			       
	} else {
	    /* do nothing: type is unknown */
	}
    }
    
    /* cleanup */
    if(result != NULL) {
	(*result) = res;
    } else {
	xmlEncResultDestroy(res);
    }
    xmlEncStateDestroy(state);    
    return(0);
}
							 
/** 
 * XMLS Enc state
 */
/** 
 * xmlEncStateCreate
 *
 *
 *
 */ 
static xmlEncStatePtr
xmlEncStateCreate(xmlEncCtxPtr ctx, int encrypt) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateCreate";
    xmlEncStatePtr state;
    
    if(ctx == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context  is null\n", 
	    func);	
#endif
 	return(NULL);	    
    }
    
    /*
     * Allocate a new xmlEncState and fill the fields.
     */
    state = (xmlEncStatePtr) xmlMalloc(sizeof(xmlEncState));
    if(state == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlEncState malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(state, 0, sizeof(xmlEncState));
        
    state->encrypt = encrypt;
    state->ctx = ctx;
    return(state);
}

/** 
 * xmlEncStateDestroy
 *
 *
 *
 */ 
static void
xmlEncStateDestroy(xmlEncStatePtr state) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateDestroy";
    if(state == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n", 
	    func);	
#endif
	return;	    
    }
    
    if(state->first != NULL) {
	xmlSecBinTransformDestroyAll((xmlSecTransformPtr)state->first);
    } else if(state->last != NULL) {
	xmlSecBinTransformDestroyAll((xmlSecTransformPtr)state->last);
    }
    
    memset(state, 0, sizeof(xmlEncState));
    xmlFree(state);
}

/**
 * xmlEncStateAddTransform
 *
 *
 *
 *
 */
static int
xmlEncStateAddTransform(xmlEncStatePtr state, xmlSecTransformPtr transform) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateAddTransform";

    if((state == NULL) || !xmlSecTransformCheckType(transform, xmlSecTransformTypeBinary)) { 
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: state is null or transform is invalid\n", 
	    func);	
#endif
	return(-1);	    
    }
    if(state->last == NULL) {
	state->first = state->last = (xmlSecBinTransformPtr)transform;
    } else if(xmlSecBinTransformAddAfter((xmlSecTransformPtr)state->last, 
					 transform) != NULL) {
	 state->last = (xmlSecBinTransformPtr)transform;
    } else {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add transform\n", 
	    func);	
#endif
	return(-1);	    
    }
    return(0);
}

/**
 * xmlEncStateAddFirstTransform
 *
 *
 *
 *
 */
static int
xmlEncStateAddFirstTransform(xmlEncStatePtr state, xmlSecTransformPtr transform) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncStateAddFirstTransform";

    if((state == NULL) || !xmlSecTransformCheckType(transform, xmlSecTransformTypeBinary)) { 
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: state is null or transform is invalid\n", 
	    func);	
#endif
	return(-1);	    
    }
    if(state->first == NULL) {
	state->first = state->last = (xmlSecBinTransformPtr)transform;
    } else if(xmlSecBinTransformAddBefore((xmlSecTransformPtr)state->first, 
					  transform) != NULL) {
	 state->first = (xmlSecBinTransformPtr)transform;
    } else {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add transform\n", 
	    func);	
#endif
	return(-1);	    
    }
    return(0);
}

/**
 * XML Enc Result
 */
/**
 * xmlEncResultCreate
 *
 *
 *
 *
 */ 		
xmlEncResultPtr		
xmlEncResultCreate(xmlEncCtxPtr ctx, int encrypt, xmlNodePtr node) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncResultCreate";
    xmlEncResultPtr result;
    
    if(ctx == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n", 
	    func);	
#endif
	return(NULL);	    
    }
    
    /*
     * Allocate a new xmlEncResult and fill the fields.
     */
    result = (xmlEncResultPtr) xmlMalloc(sizeof(xmlEncResult));
    if(result == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: xmlEncResult malloc failed\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(result, 0, sizeof(xmlEncResult));
        
    result->ctx = ctx;
    result->self = node;
    result->encrypt = encrypt;
    return(result);
}

/**
 * xmlEncResultDestoy
 *
 *
 *
 *
 */ 		
void
xmlEncResultDestroy(xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncResultDestroy";
    if(result == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: result is null\n", 
	    func);	
#endif
	return;	    
    }
    
    if(result->key != NULL) {
	xmlSecKeyDestroy(result->key, 0);
    }

    if(result->buffer != NULL) {
	xmlBufferEmpty(result->buffer);
	xmlBufferFree(result->buffer); 	
    }

    if(result->id != NULL) {
	xmlFree(result->id);
    }
    if(result->type != NULL) {
	xmlFree(result->type);
    }
    if(result->mimeType != NULL) {
	xmlFree(result->mimeType);
    }
    if(result->encoding != NULL) {
	xmlFree(result->encoding);
    }
    
    memset(result, 0, sizeof(xmlEncResult));
    xmlFree(result);
}


/**
 * xmlEncResultDebugDump
 *
 *
 *
 */
void
xmlEncResultDebugDump(FILE *output, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncResultDebugDump";

    if(result == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: result is null\n", 
	    func);	
#endif
	return;	    
    }

    if(result->encrypt) {    
        fprintf(output, "= ENCRYPTION RESULT\n");
    } else {
        fprintf(output, "= DECRYPTION RESULT (%s)\n",
	    (result->replaced) ? "replaced" : "not-replaced" );
    }
    if(result->id != NULL) {
	fprintf(output, "== Id: \"%s\"\n", result->id);
    }
    if(result->type != NULL) {
	fprintf(output, "== Type: \"%s\"\n", result->type);
    }
    if(result->mimeType != NULL) {
	fprintf(output, "== MimeType: \"%s\"\n", result->mimeType);
    }
    if(result->encoding != NULL) {
	fprintf(output, "== Encoding: \"%s\"\n", result->encoding);
    }
    
    if(result->key != NULL) {
	xmlSecDebugKeyDump(output, result->key);
    }

    if(result->buffer != NULL) {
	fprintf(output, "== start buffer:\n");
	fwrite(xmlBufferContent(result->buffer), 
	       xmlBufferLength(result->buffer), 1,
	       output);
	fprintf(output, "\n== end buffer\n");
    }	    

}


/**
 * xmlEncEncryptedDataRead
 *
 *
 */
static int
xmlEncEncryptedDataRead(xmlEncStatePtr state, xmlNodePtr encData, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncEncryptedDataRead";
    xmlNodePtr cur;
    xmlNodePtr keyInfoNode = NULL;
    xmlSecTransformPtr encryptionMethod = NULL;
    xmlSecKeyType keyType;
    int ret;
    
    if((state == NULL) || (encData == NULL) || (result == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: state or encData node is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    result->id = xmlGetProp(encData, BAD_CAST "Id");
    result->type = xmlGetProp(encData, BAD_CAST "Type");
    result->mimeType = xmlGetProp(encData, BAD_CAST "MimeType");
    result->encoding = xmlGetProp(encData, BAD_CAST "Encoding");
    cur = xmlSecGetNextElementNode(encData->children);    
        
    /* first node is optional EncryptionMethod */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "EncryptionMethod", xmlSecEncNs))) {
	encryptionMethod = xmlSecTransformNodeRead(cur, xmlSecUsageEncryptionMethod, 0);
	cur = xmlSecGetNextElementNode(cur->next);
    } else if((state->ctx != NULL) && (state->ctx->encryptionMethod != xmlSecTransformUnknown)){
	/* get encryption method from the context */
	encryptionMethod = xmlSecTransformCreate(state->ctx->encryptionMethod,
						xmlSecUsageEncryptionMethod, 0);
    } else {
#ifdef XMLSEC_DEBUG
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: EncryptionMethod not specified\n",
	    func);
#endif		
	return(-1);
    }
    if(encryptionMethod == NULL) {
#ifdef XMLSEC_DEBUG
         xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to read or create EncryptionMethod\n",
	    func);
#endif		
	return(-1);
    }    
    ret = xmlEncStateAddTransform(state, encryptionMethod);
    if(ret < 0) {    
#ifdef XMLSEC_DEBUG
         xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add EncryptionMethod\n",
	    func);
#endif		
	xmlSecTransformDestroy(encryptionMethod, 1); 
	return(-1);
    }
    xmlSecBinTransformSetEncrypt(encryptionMethod, state->encrypt);
    result->encryptionMethod = encryptionMethod->id;
    if(state->encrypt == 1) {
	keyType = xmlSecBinTransformIdGetEncKeyType(result->encryptionMethod);
    } else {
	keyType = xmlSecBinTransformIdGetDecKeyType(result->encryptionMethod);
    }
    
    /* next node is optional KeyInfo */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "KeyInfo", xmlSecDSigNs))) {
	keyInfoNode = cur;
	result->key = xmlSecKeyInfoRead(cur, encryptionMethod->id, 
				keyType, xmlEncStateGetKeysContext(state));
	cur = xmlSecGetNextElementNode(cur->next);
    }    
    if((result->key == NULL) && (xmlEncStateGetKeysContext(state) != NULL) && 
	      (xmlEncStateGetKeysContext(state)->allowedOrigins & xmlSecKeyOriginKeyManager)) {
	result->key = xmlSecKeySearcCtxFindKey(xmlEncStateGetKeysContext(state),
			encryptionMethod->id, keyType,  NULL);
	
    }    
    if(result->key == NULL) {
#ifdef XMLSEC_DEBUG    
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to find encryption key\n",
	    func);
#endif	    
	return(-1);
    }
    ret = xmlSecTransformAddKey(encryptionMethod, result->key);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG    
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add key\n",
	    func);
#endif	    
	return(-1);
    }
    if((state->encrypt == 1) && (keyInfoNode != NULL)) {
	/* update KeyInfo! */
	ret = xmlSecKeyInfoWrite(keyInfoNode, result->key, 
				 xmlSecBinTransformIdGetDecKeyType(result->encryptionMethod),
				 xmlEncStateGetKeysContext(state));
	if(ret < 0) {
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
	        "%s: failed to write \"KeyInfo\"\n",
	        func);
#endif	    
	    return(-1);
	}	
    }

    /* next is required CipherData node */
    if((cur == NULL) || (!xmlSecCheckNodeName(cur, BAD_CAST "CipherData", xmlSecEncNs))) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: required element \"CipherData\" missed\n",
	    func);
#endif	    	
	return(-1);
    }
    state->cipherDataNode = cur;
    cur = xmlSecGetNextElementNode(cur->next);

    /* next is optional EncryptionProperties node (we simply ignore it) */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "EncryptionProperties", xmlSecEncNs))) {
	cur = xmlSecGetNextElementNode(cur->next);
    }

    if(state->encrypt == 1) {
	xmlSecTransformPtr memBuf = NULL;	
	xmlSecTransformPtr base64;
	/* last transform for encryption is base64 encode */
	
	base64 = xmlSecTransformCreate(xmlSecEncBase64Encode, 0, 0);
	if(base64 == NULL) {
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create base64 encode transform\n",
		func);
#endif	    	
	    return(-1);
	}
	ret = xmlEncStateAddTransform(state, base64);
	if(ret < 0) {    
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
    		"%s: failed to add Base64 encrypt\n",
		func);
#endif		
	    xmlSecTransformDestroy(base64, 1); 
	    return(-1);
	}
	
	/* add mem buf at the end */
	memBuf = xmlSecTransformCreate(xmlSecMemBuf, 0, 0);
	if(memBuf == NULL) {
#ifdef XMLSEC_DEBUG    
    	    xmlGenericError(xmlGenericErrorContext,
    		"%s: failed to create memBuf encode transform\n",
		func);
#endif	    	
	    return(-1);
	}
	ret = xmlEncStateAddTransform(state, memBuf);
        if(ret < 0) {    
#ifdef XMLSEC_DEBUG
	    xmlGenericError(xmlGenericErrorContext,
    		"%s: failed to add Base64 encrypt\n",
		func);
#endif		
	    xmlSecTransformDestroy(memBuf, 1); 
	    return(-1);
	}
    }
    
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
	 xmlGenericError(xmlGenericErrorContext,
	    "%s: unexpected node found\n",
	    func);
#endif		
	return(-1);
    }
        
    return(0);
}

/**
 * xmlEncReplaceNode
 *
 *
 *
 *
 */
static int
xmlEncReplaceNode(xmlNodePtr node, xmlNodePtr newNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncReplaceNode";
    xmlNodePtr old;
    xmlNodePtr ptr;
    xmlNodePtr dummy;
    
    if((node == NULL) || (newNode == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: node or new node is null\n", 
	    func);	
#endif
	return(-1);	    
    }    
	    
    dummy = xmlNewDocNode(newNode->doc, NULL, BAD_CAST "dummy", NULL);
    if(dummy == NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: dummy node creation failed\n", 
	    func);	
#endif
	return(-1);
    }
	    
    if(newNode == xmlDocGetRootElement(newNode->doc)) {
	ptr = xmlDocSetRootElement(newNode->doc, dummy);
    } else {
	ptr = xmlReplaceNode(newNode, dummy);
    }
    if(ptr == NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to replace template node\n", 
	    func);	
#endif
	xmlFreeNode(dummy);
	return(-1);
    }
	    
    if(node == xmlDocGetRootElement(node->doc)) {
	old = xmlDocSetRootElement(node->doc, ptr);		
    } else {
	old = xmlReplaceNode(node, ptr);
    }
    if(old == NULL) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to replace node\n", 
	    func);	
#endif
	xmlFreeNode(ptr);
	return(-1);
    }
    xmlFreeNode(old);
    return(0);
}

/**
 * xmlEncReplaceContent
 *
 *
 *
 *
 */
static int
xmlEncReplaceContent(xmlNodePtr node, xmlNodePtr newNode) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncReplaceContent";
    xmlNodePtr dummy;
    xmlNodePtr ptr;

    if((node == NULL) || (newNode == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: node or new node is null\n", 
	    func);	
#endif
	return(-1);	    
    }    
	    
    dummy = xmlNewDocNode(newNode->doc, NULL, BAD_CAST "dummy", NULL);
    if(dummy == NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: dummy node creation failed\n", 
	    func);	
#endif
	return(-1);
    }
	    
    if(newNode == xmlDocGetRootElement(newNode->doc)) {
	ptr = xmlDocSetRootElement(newNode->doc, dummy);
    } else {
	ptr = xmlReplaceNode(newNode, dummy);
    }
    if(ptr == NULL) {
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to replace template node\n", 
	    func);	
#endif
	xmlFreeNode(dummy);
	return(-1);
    }
	    
    xmlNodeSetContent(node, NULL);
    xmlAddChild(node, ptr);
    return(0);
}


/**
 * xmlEncReplaceNodeBuffer
 *
 *
 *
 *
 */
static int
xmlEncReplaceNodeBuffer(xmlNodePtr node, const unsigned char *buffer, size_t size) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncReplaceNodeBuffer";
    static const char dummyPrefix[] = "<dummy>";
    static const char dummyPostfix[] = "</dummy>";
    xmlDocPtr doc;
    xmlNodePtr ptr1, ptr2;
    
    if((node == NULL) || (buffer == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: node is null\n", 
	    func);	
#endif
	return(-1);	    
    }

	    
    doc = xmlSecParseMemoryExt((unsigned char*)dummyPrefix, strlen(dummyPrefix),
			       buffer, size,
			       (unsigned char*)dummyPostfix, strlen(dummyPostfix));
    if(doc == NULL){
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to parse buffer\n", 
	    func);	
#endif
	return(-1);	    	
    }
	    
    ptr1 = xmlDocGetRootElement(doc);
    if(ptr1 == NULL){
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: doc root is null\n", 
	    func);	
#endif
	xmlFreeDoc(doc);
	return(-1);	    	
    }
    
    ptr1 = ptr1->children;
    while(ptr1 != NULL) {
	ptr2 = ptr1->next;
	xmlUnlinkNode(ptr1);
	xmlAddPrevSibling(node, ptr1);
	ptr1 = ptr2;
    }
	    
    xmlUnlinkNode(node);
    xmlFreeNode(node);  
    xmlFreeDoc(doc);
    return(0);
}


xmlSecKeyPtr	 
xmlEncryptedKeyNodeRead(xmlNodePtr cur, xmlSecTransformId id, 
			xmlSecKeyType type, const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncryptedKeyNodeRead";
    xmlEncCtxPtr encCtx = NULL;
    xmlSecKeyPtr key = NULL;
    xmlEncResultPtr encResult = NULL; 
    int ret;
    
    if((id == xmlSecTransformUnknown) || (cur == NULL) || (ctx == NULL)){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: id or ctx is null\n",
	    func);
#endif	    
	return(NULL);
    }
    
    /**
     * Init Enc context
     */    
    encCtx = xmlEncCtxCreate(ctx);
    if(encCtx == NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: enc ctx create failed\n",
	    func);
#endif	    
	goto done;
    }
    encCtx->ignoreType = 1; /* do not substitute the node! */
    
    ret = xmlEncDecrypt(encCtx, cur, &encResult);
    if((ret < 0) || (encResult == NULL) || (encResult->buffer == NULL)){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: node decrypt failed\n",
	    func);
#endif	    
	goto done;
    } 

    key = xmlSecKeyReadBin(xmlSecBinTransformIdGetKeyId(id), 
			   xmlBufferContent(encResult->buffer),
			   xmlBufferLength(encResult->buffer));
    if(key == NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create the key\n",
	    func);
#endif	    
	goto done;
    }			   
    
done:
    if(encResult != NULL) {
	xmlEncResultDestroy(encResult);
    }
    if(encCtx != NULL) {
	xmlEncCtxDestroy(encCtx);
    }    
    return(key);
}

int		 	
xmlEncryptedKeyNodeWrite(xmlNodePtr cur, xmlSecKeyPtr key,  xmlSecKeyType type, 
			 const xmlSecKeysReadContext *ctx) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncryptedKeyNodeWrite";
    xmlEncCtxPtr encCtx = NULL;
    unsigned char *keyBuf = NULL;
    size_t keySize = 0;
    int ret;
    int res = -1;
    
    if((cur == NULL) || (ctx == NULL) || (key == NULL)){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: id or ctx is null\n",
	    func);
#endif	    
	return(-1);
    }
    
    /**
     * Init Enc context
     */    
    encCtx = xmlEncCtxCreate(ctx);
    if(encCtx == NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: enc ctx create failed\n",
	    func);
#endif	    
	goto done;
    }
    encCtx->ignoreType = 1; /* do not substitute the node! */

    
    ret = xmlSecKeyWriteBin(key, type, &keyBuf, &keySize);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: key write failed\n",
	    func);
#endif	    
	goto done;
    }
    
    ret = xmlEncEncryptMemory(encCtx, cur, keyBuf, keySize, NULL);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: key encrypt failed\n",
	    func);
#endif	    
	goto done;
    }
    
    res = 0;

done:
    if(keyBuf != NULL) {
	memset(keyBuf, 0, keySize);
	xmlFree(keyBuf); keyBuf = NULL;
    }
    if(encCtx != NULL) {
	xmlEncCtxDestroy(encCtx);
    }    
    return(res);
}

static int
xmlEncChipherDataNodeRead(xmlEncCtxPtr ctx, xmlEncStatePtr state, 
			  xmlNodePtr node, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncChipherDataNodeRead";
    xmlNodePtr cur;
    int ret;
    
    if((state == NULL) || (ctx == NULL) || (node == NULL) || (result == NULL)){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: state, node or ctx is null\n",
	    func);
#endif	    
	return(-1);
    }
    
    cur = xmlSecGetNextElementNode(node->children);

    /* CipherValue or CipherReference */
    if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "CipherValue", xmlSecEncNs))) {
	ret = xmlEncChipherValueNodeRead(ctx, state, cur, result);
	if(ret < 0){
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read CipherValue node\n",
		func);
#endif	    
	    return(-1);
	}
	cur = xmlSecGetNextElementNode(cur->next);	
    } else if((cur != NULL) && (xmlSecCheckNodeName(cur, BAD_CAST "CipherReference",  xmlSecEncNs))) { 
	ret = xmlEncChipherRefNodeRead(ctx, state, cur, result);
	if(ret < 0){
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read CipherReference node\n",
		func);
#endif	    
	    return(-1);
	}
	cur = xmlSecGetNextElementNode(cur->next);	
    }

    if(cur != NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: found unexpected node \"%s\"\n",
	    func, cur->name);
#endif	    
	return(-1);
    }
    return(0);
}


/**
 * xmlEncCipherDataNodeWrite
 *
 *
 *
 *
 */
static int
xmlEncCipherDataNodeWrite(xmlNodePtr cipherDataNode,
		      const unsigned char *buf, size_t size) {
    static const char func[] _UNUSED_VARIABLE_ ="xmlEncCipherDataNodeWrite";
    xmlNodePtr cur; 

    if((cipherDataNode == NULL) || (buf == NULL)){
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: cipherDataNode or buf is null\n", 
	    func);	
#endif
	return(-1);	    
    }

    cur = xmlSecGetNextElementNode(cipherDataNode->children);
    if(cur == NULL) {
	cur = xmlSecAddChild(cipherDataNode, BAD_CAST "CipherValue", xmlSecEncNs);
	if(cur == NULL) {
#ifdef XMLSEC_DEBUG
	    xmlGenericError(xmlGenericErrorContext,
	        "%s: failed to create CipherValue node\n", 
		func);	
#endif
	    return(-1);	    	    
	}
        xmlNodeSetContent(cur, BAD_CAST "\n");
	xmlNodeAddContentLen(cur, buf, size);
        xmlNodeAddContent(cur, BAD_CAST "\n");
	return(0);
    } else if(xmlSecCheckNodeName(cur, BAD_CAST "CipherValue", xmlSecEncNs)) {
        xmlNodeSetContent(cur, BAD_CAST "\n");
	xmlNodeAddContentLen(cur, buf, size);
        xmlNodeAddContent(cur, BAD_CAST "\n");
	cur = xmlSecGetNextElementNode(cur->next);
    } else if(xmlSecCheckNodeName(cur, BAD_CAST "CipherReference", xmlSecEncNs)) {
	/* do nothing! */
	cur = xmlSecGetNextElementNode(cur->next);
    }
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: found unexpected node \"%s\"\n",
	    func, cur->name);
#endif	    
	return(-1);
    }
    return(0);
}


static int
xmlEncChipherValueNodeRead(xmlEncCtxPtr ctx, xmlEncStatePtr state, 
			   xmlNodePtr node, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncChipherValueNodeRead";
    xmlSecTransformPtr base64;
    xmlSecTransformPtr memBuf;
    xmlChar *content;
    int ret;
    
    if((state == NULL) || (ctx == NULL) || (node == NULL) || (result == NULL)){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: state, node or ctx is null\n",
	    func);
#endif	    
	return(-1);
    }

    /* first transform for decryption is base64 decode */	
    base64 = xmlSecTransformCreate(xmlSecEncBase64Decode, 0, 0);
    if(base64 == NULL) {
#ifdef XMLSEC_DEBUG    
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to create base64 dencode transform\n",
	    func);
#endif	    	
	return(-1);
    }

    ret = xmlEncStateAddFirstTransform(state, base64);
    if(ret < 0) {    
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add Base64 decode\n",
	    func);
#endif		
	xmlSecTransformDestroy(base64, 1); 
	return(-1);
    }


    /* add mem buf at the end */
    memBuf = xmlSecTransformCreate(xmlSecMemBuf, 0, 0);
    if(memBuf == NULL) {
#ifdef XMLSEC_DEBUG    
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to create memBuf encode transform\n",
	    func);
#endif	    	
	return(-1);
    }
    ret = xmlEncStateAddTransform(state, memBuf);
    if(ret < 0) {    
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to add Base64 encrypt\n",
	    func);
#endif		
	xmlSecTransformDestroy(memBuf, 1); 
	return(-1);
    }


    /* get node content */
    content = xmlNodeGetContent(node);
    if(content == NULL) {
#ifdef XMLSEC_DEBUG    
        xmlGenericError(xmlGenericErrorContext,
    	    "%s: failed to get node content\n",
	    func);
#endif	    	
	return(-1);
    }
	
    /* decrypt the data */
    ret = xmlSecBinTransformWrite((xmlSecTransformPtr)state->first, 
				  content, xmlStrlen(content));
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to decrypt the data\n", 
	    func);	
#endif
	xmlFree(content);
	return(-1);	    	
    }
	
    /* flush everything */
    ret = xmlSecBinTransformFlush((xmlSecTransformPtr)state->first);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize encryption\n", 
    	    func);	
#endif
	xmlFree(content);
        return(-1);	    	
    }

    result->buffer = xmlSecMemBufTransformGetBuffer((xmlSecTransformPtr)state->last, 1);
    if(result->buffer == NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to get buffer\n", 
    	    func);	
#endif
	xmlFree(content);
        return(-1);	    	
    }
    
    xmlFree(content);
    return(0);
}

static int			
xmlEncChipherRefNodeRead(xmlEncCtxPtr ctx, xmlEncStatePtr state, 
			 xmlNodePtr node, xmlEncResultPtr result) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlEncChipherRefNodeRead";
    xmlSecBinTransformPtr transform;
    xmlSecTransformStatePtr transState = NULL;
    xmlNodePtr cur;
    xmlChar *uri = NULL;
    int res = -1;
    int ret;
    
    if((state == NULL) || (ctx == NULL) || (node == NULL) || (result == NULL)){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: state, node or ctx is null\n",
	    func);
#endif	    
	return(-1);
    }
    cur = xmlSecGetNextElementNode(node->children); 
    
    uri = xmlGetProp(node, BAD_CAST "URI");
    transState = xmlSecTransformStateCreate(node->doc, NULL, (char*)uri);
    if(transState == NULL){
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create transforms state\n",
	    func);
#endif	    
	goto done;
    }	
    
    /* first is optional Transforms node */
    if((cur != NULL) && xmlSecCheckNodeName(cur, BAD_CAST "Transforms", xmlSecEncNs)) {
	ret = xmlSecTransformsNodeRead(transState, cur);
	if(ret < 0){
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read \"Transforms\"\n",
		func);
#endif	    
	    goto done;
	}	
	cur = xmlSecGetNextElementNode(cur->next);
    }
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG    
	xmlGenericError(xmlGenericErrorContext,
	    "%s: found unexpected node\n",
	    func);
#endif	    
	goto done;
    }

    /* copy all transforms from encryption state to transform state */
    while(state->first != NULL) {
	transform = state->first;
	state->first = state->first->next;
	transform->next = NULL;
	if(state->first != NULL) {
	    state->first->prev = NULL;
	}
	
	ret = xmlSecTransformStateUpdate(transState, (xmlSecTransformPtr)transform);
	if(ret < 0) {
#ifdef XMLSEC_DEBUG    
	    xmlGenericError(xmlGenericErrorContext,
		"%s: found unexpected node\n",
		func);
#endif	    
	    xmlSecBinTransformDestroyAll((xmlSecTransformPtr)transform);
	    goto done;
	}
    }
    state->last = NULL;
    ret = xmlSecTransformStateFinal(transState, xmlSecTransformResultBinary);
    if((ret < 0) || (transState->curBuf == NULL)){
#ifdef XMLSEC_DEBUG
	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to finalize transforms\n",
	    func);
#endif
	goto done;
    }
    result->buffer = transState->curBuf;
    transState->curBuf = NULL;
    
    res = 0;
done:
    if(uri != NULL) {
	xmlFree(uri);
    }
    if(transState != NULL) {
	xmlSecTransformStateDestroy(transState);
    }
    return(res);
}

#endif /* XMLSEC_NO_XMLENC */

