/** 
 * XML Security standards test: XMLEnc
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>
#include <time.h>


#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

#include <libxslt/xslt.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysmngr.h>
#include <xmlsec/xmlenc.h>


void testEncUsage(FILE *f, const char *command);

void testEncUsage(FILE *f, const char *command) {
    fprintf(stderr, "Usage:\n");
    fprintf(f, "%s encrypt|decrypt [<options>] [<tmpl1> [<tmpl2> ... ]]\n", command);
    fprintf(f, "where <options> are\n");
    fprintf(f, "== Encryption modes\n");
    fprintf(f, "  --data <datafile>      create XML Encryption for data in the file\n");
    fprintf(f, "  --node <datafile>[#<nodeId>]\n  create XML Encryption by replacing\n");
    fprintf(f, "                         the node or node content (as described in tmpl)\n");
    fprintf(f, "== Key Manager:\n");
    fprintf(f, "  --privkey[:name] <file>[,<cafile1>[,<cafile2>[...]]] the private PEM key file\n");
    fprintf(f, "                         and the list of cert file that verify it\n");
    fprintf(f, "  --pubkey[:name] <file> the public PEM key file\n");
    fprintf(f, "== Result generation/printing:\n");
    fprintf(f, "  --dump                 dump the debug verification results\n");
    fprintf(f, "== Misc.:\n");
    fprintf(f, "  --retry <number>       repeat operation with each file <number> times\n" );
}


typedef enum _testEncType {
    testEncTypeUnknown,
    testEncTypeData,
    testEncTypeNode
} testEncType;

/**
 * Init/Shutdown
 */
int			testEncInit			(int encrypt);
void			testEncShutdown			(void);

/**
 * Parsing the command line
 */
int			testEncReadKeyOrigins		(char *keyOrigins);
int			testEncReadKey			(int privateKey, 
							 char *param,
							 char *name);
int 			testEncReadKeysFile		(char *keysFile);
int			testEncReadRetries		(char *retries);







int 			testEnc				(const char* filename, 
							int encrypt,
							testEncType type,
							char *uri);

int 			testEncEncrypt			(xmlDocPtr doc,
							 testEncType type,
							 char *uri);
int 			testEncDecrypt			(xmlDocPtr doc);

void			testXmlMemory			(void);


xmlEncCtxPtr encCtx = NULL;
xmlSecSimpleKeyMngrPtr keyMgr = NULL; 
xmlSecKeysReadContext keysReadCtx;
int retryNumber = 1;

clock_t total_time = 0;
int doDump = 0;

int main(int argc, char **argv) {
    testEncType type = testEncTypeUnknown;
    char *uri = NULL;
    int encrypt = 0;    
    int ret = -1;
    int res = -1;
    int pos;
    int argsOk = 0;    
    char *name;
    
    /*
     * Parse command line
     */
    if( argc < 3 ) {
	fprintf(stderr, "Error: wrong number of arguments.\n");
	goto done;		
    } 
    if(strcmp(argv[1], "encrypt") == 0) {
	encrypt = 1;
    } else if(strcmp(argv[1], "decrypt") == 0) {
	encrypt = 0;
    } else {
        fprintf(stderr, "Error: bad mode \"%s\".\n", argv[1]);
        goto done;
    }
    
    ret = testEncInit(encrypt);
    if(ret < 0) {
	goto done;
    }
    
    /* load options */
    pos = 2;
    while(pos < argc) {
	if(argv[pos][0] != '-') {
	    /* it's filename? */
	    break;
	} else if((strcmp(argv[pos], "--data") == 0) && (pos + 1 < argc)) {	  
	    type = testEncTypeData;
	    uri = argv[++pos];
	} else if((strcmp(argv[pos], "--node") == 0) && (pos + 1 < argc)) {	  
	    type = testEncTypeNode;
	    uri = argv[++pos];
	} else if((strcmp(argv[pos], "--keys") == 0) && (pos + 1 < argc)) {	  
	    ret = testEncReadKeysFile(argv[++pos]);
	} else if((strcmp(argv[pos], "--retry") == 0) && (pos + 1 < argc)) {
	    ret = testEncReadRetries(argv[++pos]);
	} else if((strncmp(argv[pos], "--privkey", 9) == 0) && (pos + 1 < argc)) {
	    name = strchr(argv[pos], ':');
	    if(name != NULL) ++name;
	    ret = testEncReadKey(1, argv[++pos], name); 
	} else if((strncmp(argv[pos], "--pubkey", 8) == 0) && (pos + 1 < argc)) {
	    name = strchr(argv[pos], ':');
	    if(name != NULL) ++name;
	    ret = testEncReadKey(0, argv[++pos], name); 
	} else if((strcmp(argv[pos], "--keys-origins") == 0) && (pos + 1 < argc)){
	    ret = testEncReadKeyOrigins(argv[++pos]);
	} else if(strcmp(argv[pos], "--dump") == 0) {
	    doDump = 1;
	} else {
	    fprintf(stderr, "Error: bad option \"%s\".\n", argv[pos]);
	    goto done;
	}
	if(ret < 0) {
	    goto done;
	}
	++pos;
    }
    argsOk = 1;

    while(pos < argc) { 
	res = testEnc(argv[pos++], encrypt, type, uri);	    
	if(res < 0) {	
	    break;
	}
    }
    fprintf(stderr, "Executed %d tests in %ld msec\n", retryNumber, total_time / (CLOCKS_PER_SEC / 1000));

done:
    if(!argsOk) {
	testEncUsage(stderr, argv[0]);
    }
        
    testEncShutdown();
    return((res >= 0) ? 0 : 1);
}

int 
testEnc(const char* filename, int encrypt, testEncType type, char *uri) {
    xmlDocPtr doc;
    int ret = 0;
    int i;
    clock_t start_time;
    
    /*
     * build an XML tree from a the file; we need to add default
     * attributes and resolve all character and entities references
     */
    xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
    xmlSubstituteEntitiesDefault(1);


    for(i = 0; i < retryNumber; ++i) {
        doc = xmlParseFile(filename);
	if (doc == NULL) {
	    fprintf(stderr, "Error: unable to parse file \"%s\"\n", filename);
	    return(-1);
	}
    
	/*
	 * Check the document is of the right kind
	 */    
	if(xmlDocGetRootElement(doc) == NULL) {
    	    fprintf(stderr,"Error: empty document for file \"%s\"\n", filename);
	    xmlFreeDoc(doc);
	    return(-1);
	}
        
	/*
	 * encrypt/decrypt encryptature
	 */      
	start_time = clock();
	if(encrypt) {
	    fprintf(stderr, "Encrypting tmpl \"%s\"\n", filename);	
	    ret = testEncEncrypt(doc, type, uri);
	} else {
	    fprintf(stderr, "Decrypting file \"%s\"\n", filename);	
	    ret = testEncDecrypt(doc);
	}
	total_time += clock() - start_time;

	if(ret < 0) {
	    xmlFreeDoc(doc);    
	    break;
	}
	
	xmlFreeDoc(doc);    
    }    
    
    return(ret);
}
	
int testEncEncrypt(xmlDocPtr tmpl, testEncType type, char *uri) {    
    xmlEncResultPtr encResult = NULL;
    xmlChar *result = NULL;	
    xmlDocPtr doc = NULL;
    int len;
    int ret;
    int res = -1;

    if((type == testEncTypeData) && (uri != NULL)) {
	xmlBufferPtr buf = NULL;
	FILE *f;
	unsigned char tmp[1024];
	
	buf = xmlBufferCreate();
	if(buf == NULL) {
    	    fprintf(stderr,"Error: failed to create buffer\n");
	    goto done;    
	}
	
	f = fopen(uri, "r");
	if(f == NULL) {
    	    fprintf(stderr,"Error: failed to open file \"%s\" (%d)\n", uri, errno);
	    xmlBufferFree(buf);
	    goto done;    
	}
	while((ret = fread(tmp, 1, sizeof(tmp), f)) > 0) {
	    xmlBufferAdd(buf, tmp, ret);
	}

	if(ret < 0){
    	    fprintf(stderr,"Error: failed to read file \"%s\" (%d)\n", uri, errno);
	    fclose(f);
	    xmlBufferFree(buf);
	    goto done;    
	}
	fclose(f);
	
	ret = xmlEncEncryptMemory(encCtx, xmlDocGetRootElement(tmpl), 
			          xmlBufferContent(buf), xmlBufferLength(buf), 
				  &encResult);
	if(ret < 0) {
    	    fprintf(stderr,"Error: xmlEncEncryptMemory() failed \n");
	    xmlBufferFree(buf);
	    goto done;    
	} 
	xmlBufferFree(buf);
    } else if((type == testEncTypeNode) && (uri != NULL)) { 
	char *id;
	xmlNodePtr cur;
	
	id = strchr(uri, '#');
	if(id != NULL) {
	    *(id++) = '\0';
	}
        
	/** 
	 * Load doc
	 */
	doc = xmlParseFile(uri);
	if (doc == NULL) {
	    fprintf(stderr, "Error: unable to parse file \"%s\"\n", uri);
	    goto done;    
	}

	/**
	 * What do we want to replace?
	 */    
	if(id == NULL) {
	    cur = xmlDocGetRootElement(doc);
	} else {
	    cur = xmlSecFindNodeById(xmlDocGetRootElement(doc), BAD_CAST id);
	}
	
	/*
	 * Did we found node?
	 */    
	if(cur == NULL) {
    	    fprintf(stderr,"Error: empty document for file \"%s\" or unable to find node \"%s\"\n", uri, (id != NULL) ? id : "NULL");
	    goto done;    
	}
	
	ret = xmlEncEncryptXmlNode(encCtx, xmlDocGetRootElement(tmpl),
				cur, &encResult);	
	if(ret < 0) {
    	    fprintf(stderr,"Error: xmlEncEncryptXmlNode() failed \n");
	    goto done;    
	} 
    } else {
        fprintf(stderr,"Error: unknown type or bad type parameters\n");
	goto done;    
    }
    	
    /*
     * Print document out in default UTF-8 encoding
     */
    if((encResult != NULL) && (encResult->replaced) && (doc != NULL)) {
	xmlDocDumpMemoryEnc(doc, &result, &len, NULL);  
    } else {
	xmlDocDumpMemoryEnc(tmpl, &result, &len, NULL);
    }
    if(result == NULL) {
        fprintf(stderr,"Error: failed to dump document to memory\n");
	goto done;
    }
    fwrite(result, len, 1, stdout);
    res = 0;

    if(doDump) {
    	xmlEncResultDebugDump(stderr, encResult);
    }	
        
done:    
    if(doc != NULL) {
	xmlFreeDoc(doc);
    }
    if(result != NULL) {
	xmlFree(result);        
    }
    if(encResult != NULL) {
	xmlEncResultDestroy(encResult);
    }
    return(res);
}

int testEncDecrypt(xmlDocPtr doc) {    
    xmlEncResultPtr encResult = NULL;
    xmlNodePtr cur;
    int ret;
    int res = -1;

    cur = xmlSecFindNode(xmlDocGetRootElement(doc), BAD_CAST "EncryptedData", xmlSecEncNs);
    if(cur == NULL) {
        fprintf(stderr,"Error: unable to find EncryptedData node\n");
	goto done;    
    }
    
    ret = xmlEncDecrypt(encCtx, cur, &encResult);
    if(ret < 0) {
        fprintf(stderr,"Error: xmlEncDecrypt() failed \n");
	goto done;    
    } 

    if((encResult != NULL) && encResult->replaced && (encResult->buffer != NULL)) {
	ret = xmlDocDump(stdout, doc);    
    } else if((encResult != NULL) && !encResult->replaced) {
	ret = fwrite(xmlBufferContent(encResult->buffer), 
	       xmlBufferLength(encResult->buffer),
	       1, stdout);		       
    } else {
        fprintf(stderr,"Error: bad results \n");
	goto done;    
    }

    if(ret < 0) {
        fprintf(stderr,"Error: failed to print out the result \n");
	goto done;    
    }
    
    if(doDump) {
    	xmlEncResultDebugDump(stderr, encResult);
    }	
    res = 0;
    
done:    
    if(encResult != NULL) {
	xmlEncResultDestroy(encResult);
    }
    return(res);
}

/**
 * Parsing the command line
 */
int			
testEncReadKeyOrigins(char *keyOrigins) {
    xmlSecKeyOrigin res = xmlSecKeyOriginDefault;
    char *p;
    
    p = strtok(keyOrigins, ",");
    while(p != NULL) {
	if(strcmp(p, "keymanager") == 0) {
	    res |= xmlSecKeyOriginKeyManager;
	} else if(strcmp(p, "keyname") == 0) {
	    res |= xmlSecKeyOriginKeyManager | xmlSecKeyOriginKeyName;
	} else if(strcmp(p, "keyvalue") == 0) {
	    res |= xmlSecKeyOriginKeyValue;
	} else if(strcmp(p, "retrieval-doc") == 0) {
	    res |= xmlSecKeyOriginRetrievalDocument;
	} else if(strcmp(p, "retrieval-remote") == 0) {
	    res |= xmlSecKeyOriginRetrievalRemote;
	} else if(strcmp(p, "x509") == 0) {
	    res |= xmlSecKeyOriginX509;
	} else if(strcmp(p, "pgp") == 0) {
	    res |= xmlSecKeyOriginPGP;
	} else {
	    fprintf(stderr, "Error: unknown key origin: \"%s\" (ignored)\n", p);
	    return(-1);
	}
	p = strtok(NULL, ",");
    }    
    keysReadCtx.allowedOrigins = res;
    return(0);
}

int
testEncReadKey	(int privateKey, char *param, char *name) {
    char *p;
    xmlSecKeyPtr key;
    int ret;
    
    p = strtok(param, ","); 
    if(privateKey) {
	key = xmlSecSimpleKeyMngrLoadPrivateKey(keyMgr, p, NULL, NULL);
    } else {
	key = xmlSecSimpleKeyMngrLoadPublicKey(keyMgr, p, NULL, NULL);
    }
    if(key == NULL) {
	fprintf(stderr, "Error: failed to load key from \"%s\"\n", p);
	return(-1);
    }
    if(name != NULL) {
	key->name = xmlStrdup(BAD_CAST name);
    }
    
    p = strtok(NULL, ",");
    while((p != NULL) && (privateKey)) {
	ret = xmlSecSimpleKeyMngrAddCertToKey(key, p);
	if(ret < 0){
	    fprintf(stderr, "Error: failed to load cert from \"%s\"\n", p);
	    return(-1);
	}
	p = strtok(NULL, ","); 
    }
    /* TODO: */
    return(0);
}

int
testEncReadKeysFile(char *keysFile) {
    int ret;
    
    ret = xmlSecSimpleKeyMngrLoad(keyMgr, keysFile, 0);
    if(ret < 0) {
	fprintf(stderr, "Error: failed to load keys from \"%s\".\n", keysFile);
	return(-1);
    }
    return(0);
}

int
testEncReadRetries(char *retries) {
    
    if(sscanf(retries, "%d", &retryNumber) != 1) {
	fprintf(stderr, "Error: bad number \"%s\".\n", retries);
	return(-1);
    }	    
    return(0);
}

/**
 * Init/Shutdown
 */
int 
testEncInit(int encrypt) {
    int rnd_seed = 0;

    /* 
     * Init OpenSSL
     */
    OpenSSL_add_all_algorithms();
    ERR_load_crypto_strings(); 
    while (RAND_status() != 1) {
	RAND_seed(&rnd_seed, sizeof(rnd_seed));
    }

    /*
     * Init libxml
     */     
/*
    testXmlMemory(); 
*/    
    xmlInitParser();
    LIBXML_TEST_VERSION
    
    /*
     * Init xmlsec
     */
    xmlSecInit();    


    /** 
     * Create Keys managers
     */
    keyMgr = xmlSecSimpleKeyMngrCreate();    
    if(keyMgr == NULL) {
	fprintf(stderr, "Error: failed to create keys manager\n");
	return(-1);
    }
    memset(&keysReadCtx, 0, sizeof(keysReadCtx));

    keysReadCtx.allowedOrigins = xmlSecKeyOriginAll; /* by default all keys are accepted */
    keysReadCtx.maxRetrievals = 1; /* one level retrievals only */
    keysReadCtx.findKeyCallback = xmlSecSimpleKeyMngrFindKey;
    keysReadCtx.findKeyContext = keyMgr;
    if(!encrypt) {
	keysReadCtx.maxEncKeyLevels = 1;
    }        

    /**
     * Init Enc context
     */    
    encCtx = xmlEncCtxCreate(&keysReadCtx);
    if(encCtx == NULL) {
        fprintf(stderr,"Error: failed to create Enc context\n");
	return(-1);
    }

    return(0);    
}

void
testEncShutdown(void) {
    /* destroy xmlsec objects */
    if(keyMgr != NULL) {
	xmlSecSimpleKeyMngrDestroy(keyMgr);
    }
    if(encCtx != NULL) {
	xmlEncCtxDestroy(encCtx);
    }

    xmlSecShutdown();

    xsltCleanupGlobals();
                
    /* 
     * Shutdown libxml
     */
    xmlCleanupParser();
    xmlMemoryDump();

    /**
     * Shutdown OpenSSL
     */    
    RAND_cleanup();
    ERR_free_strings();
    EVP_cleanup();
    X509_TRUST_cleanup();
}

/**************************************************************************/

xmlFreeFunc freeFunc = NULL;
xmlMallocFunc mallocFunc = NULL;
xmlReallocFunc reallocFunc = NULL;
xmlStrdupFunc strdupFunc = NULL;

static void myFreeFunc(void *mem) {
    static size_t n = 0;
    if(freeFunc) {
	fprintf(stderr, "%p: free (%d)\n", mem, ++n);
	freeFunc(mem);
    }
}

static void *myMallocFunc(size_t size) {
    static size_t n = 0;
    
    if(mallocFunc) {
	void *mem;
	
	mem = mallocFunc(size);
	fprintf(stderr, "%p: malloc (%d:%d)\n", mem, ++n, size);
	if(size == 34) {
	    fprintf(stderr, "%p: hi!\n", mem);
	}
	return(mem);
    }
    return(NULL);
}

static void *myReallocFunc(void *mem, size_t size) {
    static size_t n = 0;
    if(reallocFunc) {
	void *p;

	fprintf(stderr, "%p: realloc-in (%d:%d)\n", mem, ++n, size);	
	p = reallocFunc(mem, size);
	fprintf(stderr, "%p: realloc-out (%d:%d)\n", p, ++n, size);	
	return(p);
    }
    return(NULL);
}

static char *myStrdupFunc(const char *str) {
    static size_t n = 0;
    if(strdupFunc) {
	char *mem;
	
	mem = strdupFunc(str);
	fprintf(stderr, "%p: strdup (%d:%d)\n", mem, ++n, strlen(str) + 1); 
	return(mem);
    }
    return(NULL);
}


void	
testXmlMemory(void) {
    xmlMemGet(&freeFunc, &mallocFunc, &reallocFunc, &strdupFunc);
    xmlMemSetup(myFreeFunc, myMallocFunc, myReallocFunc, myStrdupFunc);
}


