/** 
 * XML Security standards test: keys managemenet
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <string.h>

#include <openssl/rand.h>
#include <openssl/err.h>

#include <libxml/tree.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysmngr.h>
#include <xmlsec/transforms.h>

static const char usage[] = "<mode> [<options>] <filename> \n"
			    "where <mode> is one of following\n"
			    "  privatekeys  \t to dump keys to a file\n"
			    "  publickeys   \t to dump keys to a file\n"
			    "  createkeys   \t to create keys to a file\n"
			    "and <options> are from following list\n"
			    "  --keys <keys>\t to load keys from file <filename>\n"
			    "";
int 
test_create_keys(const char* filename);

xmlSecSimpleKeyMngrPtr keyMgr = NULL;

int main(int argc, char **argv) {
    int keysLoaded = 0;    
    int ret = 0;
    int pos;
    int rnd_seed = 0;
    
    while (RAND_status() != 1) {
	RAND_seed(&rnd_seed, sizeof(rnd_seed));
    }

        
    /*
     * Init libxml
     */     
    xmlInitParser();
    LIBXML_TEST_VERSION
    
    xmlSecInit();
    
    keyMgr = xmlSecSimpleKeyMngrCreate();
    if(keyMgr == NULL) {
	fprintf(stderr, "Error: failed to create keys manager\n");
	goto done;	
    }
    
    /*
     * Parse command line and process file
     */
    if( argc < 3 ) {
	fprintf(stderr, "Error: wrong number of arguments.\nUsage: %s %s\n", argv[0], usage);
	goto done;		
    } 
    
    /* load options */
    pos = 2;
    while(pos < argc) {
	if(argv[pos][0] != '-') {
	    /* it's filename? */
	    break;
	} else if(strcmp(argv[pos], "--keys") == 0) {	  
	    /* load keys! */
	    if((++pos) >= argc) {
		fprintf(stderr, "Error: option \"%s\" requires argument.\nUsage: %s %s\n", argv[pos-1], argv[0], usage);
		goto done;
	    }
	    
	    ret = xmlSecSimpleKeyMngrLoad(keyMgr, argv[pos], 0);
	    if(ret < 0) {
		fprintf(stderr, "Error: failed to load keys from \"%s\".\nUsage: %s %s\n", argv[pos], argv[0], usage);
		goto done;
	    }
	    keysLoaded = 1;
	} else {
	    fprintf(stderr, "Error: bad option \"%s\".\nUsage: %s %s\n", argv[pos], argv[0], usage);
	    ret = -1;
	    goto done;
	}
	++pos;
    }
    
    while(pos < argc && ret >= 0) { 
	if(strcmp(argv[1], "privatekeys") == 0) {
	    if(!keysLoaded) {
		fprintf(stderr, "Error: keys not loaded.\nUsage: %s %s\n", argv[0], usage);
		goto done;
	    }
	    ret = xmlSecSimpleKeyMngrSave(keyMgr, argv[pos], xmlSecKeyTypePrivate);
	    if(ret < 0) {
		fprintf(stderr, "Error: \"%s\" failed\n", argv[1]);
	    }
	    goto done;
	} else if(strcmp(argv[1], "publickeys") == 0) {
	    if(!keysLoaded) {
		fprintf(stderr, "Error: keys not loaded.\nUsage: %s %s\n", argv[0], usage);
		goto done;
	    }
	    ret = xmlSecSimpleKeyMngrSave(keyMgr, argv[pos], xmlSecKeyTypePublic);
	    if(ret < 0) {
		fprintf(stderr, "Error: \"%s\" failed\n", argv[1]);
	    }
	    goto done;
	} else if(strcmp(argv[1], "createkeys") == 0) {
	    ret = test_create_keys(argv[pos]);
	} else {
	    fprintf(stderr, "Error: bad mode \"%s\".\nUsage: %s %s\n", argv[1], argv[0], usage);
	    ret = -1;
	    goto done;
	}
	++pos;
    }


done:
    if(keyMgr != NULL) {
	xmlSecSimpleKeyMngrDestroy(keyMgr);
    }
    
    /* 
     * Shutdown libxml
     */
    xmlCleanupParser();
    xmlMemoryDump();
    
    RAND_cleanup();
    ERR_clear_error();

    return((ret >= 0) ? 0 : 1);
}


int 
test_create_keys(const char* filename) {
    xmlSecKeyPtr key;    
    int ret;
    
#ifndef XMLSEC_NO_HMAC    
    /* HMAC */    
    key = xmlSecKeyCreate(xmlSecHmacKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create hmac key\n"); 
	return(-1);
    }        
    ret = xmlSecHmacKeyGenerate(key, (unsigned char*)"secret", 6);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1); 
	fprintf(stderr, "Error: failed to set hmac key params\n"); 
	return(-1);
    }
    key->name = xmlStrdup(BAD_CAST "test-hmac-sha1");
    
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add hmac key\n"); 
	return(-1);		
    }
#endif /* XMLSEC_NO_HMAC */    

#ifndef XMLSEC_NO_DSA
    /* DSA */
    key = xmlSecKeyCreate(xmlSecDsaKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create dsa key\n"); 
	return(-1);
    }        
    ret = xmlSecDsaKeyGenerate(key, NULL);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1); 	
	fprintf(stderr, "Error: failed to set dsa key params\n"); 
	return(-1);
    }    
    key->name = xmlStrdup(BAD_CAST "test-dsa");
    
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add dsa key\n"); 
	return(-1);
    }
#endif /* XMLSEC_NO_DSA */    

#ifndef XMLSEC_NO_RSA
    /* RSA */
    key = xmlSecKeyCreate(xmlSecRsaKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create rsa key\n"); 
	return(-1);
    }        
    ret = xmlSecRsaKeyGenerate(key, NULL);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1); 	
	fprintf(stderr, "Error: failed to set rsa key params\n"); 
	return(-1);
    }    
    key->name = xmlStrdup(BAD_CAST "test-rsa");
    
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add rsa key\n"); 
	return(-1);
    }
#endif /* XMLSEC_NO_RSA */    

#ifndef XMLSEC_NO_DES
    /* DES */    
    key = xmlSecKeyCreate(xmlSecDesKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create des key\n"); 
	return(-1);
    }        
    ret = xmlSecDesKeyGenerate(key, NULL, 24);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1); 	
	fprintf(stderr, "Error: failed to set des key params\n"); 
	return(-1);
    }    
    key->name = xmlStrdup(BAD_CAST "test-des");

    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add des key\n"); 
	return(-1);
    }
#endif /* XMLSEC_NO_DES */

#ifndef XMLSEC_NO_AES
    /* AES 128 */    
    key = xmlSecKeyCreate(xmlSecAesKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create aes 128 key\n"); 
	return(-1);
    }        
    ret = xmlSecAesKeyGenerate(key, NULL, 128);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);  
	fprintf(stderr, "Error: failed to create aes 128 key\n"); 
	return(-1);
    }    
    key->name = xmlStrdup(BAD_CAST "test-aes128");
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add aes 128 key\n"); 
	return(-1);
    }

    /* AES 192 */    
    key = xmlSecKeyCreate(xmlSecAesKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create aes 192 key\n"); 
	return(-1);
    }        
    ret = xmlSecAesKeyGenerate(key, NULL, 192);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);  
	fprintf(stderr, "Error: failed to create aes 192 key\n"); 
	return(-1);
    }    
    key->name = xmlStrdup(BAD_CAST "test-aes192");
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add aes 192 key\n"); 
	return(-1);
    }

    /* AES 256 */    
    key = xmlSecKeyCreate(xmlSecAesKey, xmlSecKeyOriginDefault);
    if(key == NULL) {
	fprintf(stderr, "Error: failed to create aes 256 key\n"); 
	return(-1);
    }        
    ret = xmlSecAesKeyGenerate(key, NULL, 256);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);  
	fprintf(stderr, "Error: failed to create aes 256 key\n"); 
	return(-1);
    }    
    key->name = xmlStrdup(BAD_CAST "test-aes256");
    ret = xmlSecSimpleKeyMngrAddKey(keyMgr, key);
    if(ret < 0) {
	xmlSecKeyDestroy(key, 1);
	fprintf(stderr, "Error: failed to add aes 256 key\n"); 
	return(-1);
    }

#endif /* XMLSEC_NO_AES */

    ret = xmlSecSimpleKeyMngrSave(keyMgr,  filename, 
				  xmlSecKeyTypePublic | xmlSecKeyTypePrivate);
    if(ret < 0) {
	fprintf(stderr, "Error: failed to write to \"%s\"\n", filename);
	return(-1);	
    }
    
    return(0);
}    
